"""This module is strongly inspired by / copied from xarray/plot/plot.py."""

import functools

import numpy as np

from xugrid.constants import FloatDType
from xugrid.plot.utils import (
    _add_colorbar,
    _easy_facetgrid,
    _ensure_plottable,
    _process_cmap_cbar_kwargs,
    _update_axes,
    get_axis,
    label_from_attrs,
)
from xugrid.ugrid.connectivity import close_polygons

NODE = 0
EDGE = 1
FACE = 2
COORDS = [
    "node_coordinates",
    "edge_coordinates",
    "face_coordinates",
]


def get_ugrid_dim(grid, da) -> int:
    dim = da.dims[0]
    if dim == grid.node_dimension:
        return NODE
    elif dim == grid.edge_dimension:
        return EDGE
    elif dim == grid.face_dimension:
        return FACE
    else:
        allowed_dims = [grid.node_dimension, grid.edge_dimension, grid.face_dimension]
        raise ValueError(
            f"Not a valid UGRID dimension: {dim},should be one of: {allowed_dims}"
        )


def override_signature(f):
    def wrapper(func):
        func.__wrapped__ = f

        return func

    return wrapper


def _plot2d(plotfunc):
    """
    Decorator for common 2d plotting logic
    Also adds the 2d plot method to class _PlotMethods
    """  # noqa: D401
    commondoc = """
    Parameters
    ----------
    topology : Union[Ugrid1d, Ugrid2d, Tuple[FloatArray, FloatArray], Triangulation]
        Mesh topology.
    darray : DataArray
        Must be two-dimensional, unless creating faceted plots.
    figsize : tuple, optional
        A tuple (width, height) of the figure in inches.
        Mutually exclusive with ``size`` and ``ax``.
    aspect : scalar, optional
        Aspect ratio of plot, so that ``aspect * size`` gives the *width* in
        inches. Only used if a ``size`` is provided.
    size : scalar, optional
        If provided, create a new figure for the plot with the given size:
        *height* (in inches) of each plot. See also: ``aspect``.
    ax : matplotlib axes object, optional
        Axes on which to plot. By default, use the current axes.
        Mutually exclusive with ``size`` and ``figsize``.
    row : string, optional
        If passed, make row faceted plots on this dimension name.
    col : string, optional
        If passed, make column faceted plots on this dimension name.
    col_wrap : int, optional
        Use together with ``col`` to wrap faceted plots.
    xticks, yticks : array-like, optional
        Specify tick locations for *x*- and *y*-axis.
    xlim, ylim : array-like, optional
        Specify *x*- and *y*-axis limits.
    xincrease : None, True, or False, optional
        Should the values on the *x* axis be increasing from left to right?
        If ``None``, use the default for the Matplotlib function.
    yincrease : None, True, or False, optional
        Should the values on the *y* axis be increasing from top to bottom?
        If ``None``, use the default for the Matplotlib function.
    add_colorbar : bool, optional
        Add colorbar to axes.
    add_labels : bool, optional
        Use xarray metadata to label axes.
    norm : matplotlib.colors.Normalize, optional
        If ``norm`` has ``vmin`` or ``vmax`` specified, the corresponding
        kwarg must be ``None``.
    vmin, vmax : float, optional
        Values to anchor the colormap, otherwise they are inferred from the
        data and other keyword arguments. When a diverging dataset is inferred,
        setting one of these values will fix the other by symmetry around
        ``center``. Setting both values prevents use of a diverging colormap.
        If discrete levels are provided as an explicit list, both of these
        values are ignored.
    cmap : matplotlib colormap name or colormap, optional
        The mapping from data values to color space. If not provided, this
        will be either be ``'viridis'`` (if the function infers a sequential
        dataset) or ``'RdBu_r'`` (if the function infers a diverging dataset).
        See :doc:`Choosing Colormaps in Matplotlib <matplotlib:tutorials/colors/colormaps>`
        for more information.
        If *seaborn* is installed, ``cmap`` may also be a
        `seaborn color palette <https://seaborn.pydata.org/tutorial/color_palettes.html>`_.
        Note: if ``cmap`` is a seaborn color palette and the plot type
        is not ``'contour'`` or ``'contourf'``, ``levels`` must also be specified.
    colors : str or array-like of color-like, optional
        A single color or a sequence of colors. If the plot type is not ``'contour'``
        or ``'contourf'``, the ``levels`` argument is required.
    center : float, optional
        The value at which to center the colormap. Passing this value implies
        use of a diverging colormap. Setting it to ``False`` prevents use of a
        diverging colormap.
    robust : bool, optional
        If ``True`` and ``vmin`` or ``vmax`` are absent, the colormap range is
        computed with 2nd and 98th percentiles instead of the extreme values.
    extend : {'neither', 'both', 'min', 'max'}, optional
        How to draw arrows extending the colorbar beyond its limits. If not
        provided, ``extend`` is inferred from ``vmin``, ``vmax`` and the data limits.
    levels : int or array-like, optional
        Split the colormap (``cmap``) into discrete color intervals. If an integer
        is provided, "nice" levels are chosen based on the data range: this can
        imply that the final number of levels is not exactly the expected one.
        Setting ``vmin`` and/or ``vmax`` with ``levels=N`` is equivalent to
        setting ``levels=np.linspace(vmin, vmax, N)``.
    infer_intervals : bool, optional
        Only applies to pcolormesh. If ``True``, the coordinate intervals are
        passed to pcolormesh. If ``False``, the original coordinates are used
        (this can be useful for certain map projections). The default is to
        always infer intervals, unless the mesh is irregular and plotted on
        a map projection.
    subplot_kws : dict, optional
        Dictionary of keyword arguments for Matplotlib subplots. Only used
        for 2D and faceted plots.
        (see :py:meth:`matplotlib:matplotlib.figure.Figure.add_subplot`).
    cbar_ax : matplotlib axes object, optional
        Axes in which to draw the colorbar.
    cbar_kwargs : dict, optional
        Dictionary of keyword arguments to pass to the colorbar
        (see :meth:`matplotlib:matplotlib.figure.Figure.colorbar`).
    **kwargs : optional
        Additional keyword arguments to wrapped Matplotlib function.
    Returns
    -------
    artist :
        The same type of primitive artist that the wrapped Matplotlib
        function returns.
    """

    # Build on the original docstring
    plotfunc.__doc__ = f"{plotfunc.__doc__}\n{commondoc}"

    # plotfunc and newplotfunc have different signatures:
    # - plotfunc: (x, y, z, ax, **kwargs)
    # - newplotfunc: (darray, x, y, **kwargs)
    # where plotfunc accepts numpy arrays, while newplotfunc accepts a DataArray
    # and variable names. newplotfunc also explicitly lists most kwargs, so we
    # need to shorten it
    def signature(topology, darray, **kwargs):
        pass  # pragma: no-cover

    @override_signature(signature)
    @functools.wraps(plotfunc)
    def newplotfunc(
        topology,
        darray=None,
        figsize=None,
        size=None,
        aspect=None,
        ax=None,
        row=None,
        col=None,
        col_wrap=None,
        xincrease=True,
        yincrease=True,
        add_colorbar=None,
        add_labels=True,
        vmin=None,
        vmax=None,
        cmap=None,
        center=None,
        robust=False,
        extend=None,
        levels=None,
        colors=None,
        subplot_kws=None,
        cbar_ax=None,
        cbar_kwargs=None,
        xticks=None,
        yticks=None,
        xlim=None,
        ylim=None,
        norm=None,
        **kwargs,
    ):
        # All 2d plots in xarray share this function signature.
        # Method signature below should be consistent.

        # Decide on a default for the colorbar before facetgrids
        if add_colorbar is None:
            add_colorbar = True
            if (
                darray is None
                or plotfunc.__name__ == "contour"
                or (plotfunc.__name__ == "surface" and cmap is None)
            ):
                add_colorbar = False

        if subplot_kws is None:
            subplot_kws = {}

        if plotfunc.__name__ == "surface" and not kwargs.get("_is_facetgrid", False):
            if ax is None:
                # TODO: Importing Axes3D is no longer necessary in matplotlib >= 3.2.
                # Remove when minimum requirement of matplotlib is 3.2:
                from mpl_toolkits.mplot3d import Axes3D  # type: ignore  # noqa: F401

                # delete so it does not end up in locals()
                del Axes3D

                # Need to create a "3d" Axes instance for surface plots
                subplot_kws["projection"] = "3d"

            # In facet grids, shared axis labels don't make sense for surface plots
            sharex = False
            sharey = False

        # Handle facetgrids first
        if row or col:
            if darray is None:
                raise ValueError(
                    "Cannot create facetgrid with only topology and no data."
                )
            allargs = locals().copy()
            del allargs["darray"]
            del allargs["imshow_rgb"]
            allargs.update(allargs.pop("kwargs"))
            # Need the decorated plotting function
            allargs["plotfunc"] = globals()[plotfunc.__name__]
            return _easy_facetgrid(darray, kind="dataarray", **allargs)

        # For 3d plot, ensure given ax is a Axes3D object
        if (
            plotfunc.__name__ == "surface"
            and not kwargs.get("_is_facetgrid", False)
            and ax is not None
        ):
            import mpl_toolkits  # type: ignore

            if not isinstance(ax, mpl_toolkits.mplot3d.Axes3D):
                raise ValueError(
                    "If ax is passed to surface(), it must be created with "
                    'projection="3d"'
                )

        # darray may be None when plotting just edges (the mesh)
        if darray is not None:
            _ensure_plottable(darray.to_numpy())
            cmap_params, cbar_kwargs = _process_cmap_cbar_kwargs(
                plotfunc,
                darray,
                **locals(),
                _is_facetgrid=kwargs.pop("_is_facetgrid", False),
            )
        else:
            cmap_params = {}
            cbar_kwargs = {}

        if "contour" in plotfunc.__name__:
            # extend is a keyword argument only for contour and contourf, but
            # passing it to the colorbar is sufficient for imshow and
            # pcolormesh
            kwargs["extend"] = cmap_params["extend"]
            kwargs["levels"] = cmap_params["levels"]
            # if colors == a single color, matplotlib draws dashed negative
            # contours. we lose this feature if we pass cmap and not colors
            if isinstance(colors, str):
                cmap_params["cmap"] = None
                kwargs["colors"] = colors

        if "imshow" == plotfunc.__name__ and isinstance(aspect, str):
            # forbid usage of mpl strings
            raise ValueError(
                "plt.imshow's `aspect` string kwarg is not available in xugrid. "
                "Use a float instead."
            )

        for key in ["cmap", "vmin", "vmax", "norm"]:
            kwargs[key] = cmap_params.get(key)

        ax = get_axis(figsize, size, aspect, ax, **subplot_kws)
        primitive = plotfunc(
            topology,
            darray,
            ax=ax,
            **kwargs,
        )

        if size is not None:
            # Try to get a 1:1 ratio between x and y coordinates by default. If
            # colorbar is present, we need to make room for it in the
            # x-direction.  1.26 is the magic number; the colorbar takes up 26%
            # additional space by default.
            if aspect is None:
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                aspect = (xmax - xmin) / (ymax - ymin)
                if add_colorbar:
                    aspect *= 1.26

            # Preserve height, adjust width if needed; Do not call
            # ax.set_aspect: this shrinks or grows the ax relative to the
            # colobar
            figsize = (size * aspect, size)
            ax.figure.set_size_inches(figsize)

        # Label the plot with metadata
        if darray is not None and add_labels:
            # TODO: grab x and y information from topology?
            ax.set_xlabel("x")
            ax.set_ylabel("y")
            ax.set_title(darray._title_for_slice())
            if plotfunc.__name__ == "surface":
                ax.set_zlabel(label_from_attrs(darray))

        if add_colorbar:
            if add_labels and "label" not in cbar_kwargs:
                cbar_kwargs["label"] = label_from_attrs(darray)
            cbar = _add_colorbar(primitive, ax, cbar_ax, cbar_kwargs, cmap_params)
        elif cbar_ax is not None or cbar_kwargs:
            # inform the user about keywords which aren't used
            raise ValueError(
                "cbar_ax and cbar_kwargs can't be used with add_colorbar=False."
            )

        # origin kwarg overrides yincrease
        if "origin" in kwargs:
            yincrease = None

        # Spatial x and y coordinates: no need for e.g. logarithm axes.
        xscale = None
        yscale = None
        _update_axes(
            ax, xincrease, yincrease, xscale, yscale, xticks, yticks, xlim, ylim
        )

        return primitive

    return newplotfunc


@_plot2d
def scatter(grid, da, ax, **kwargs):
    dim = get_ugrid_dim(grid, da)
    x, y = getattr(grid, COORDS[dim]).T
    primitive = ax.scatter(x, y, c=da.to_numpy().ravel(), **kwargs)
    return primitive


@_plot2d
def tripcolor(grid, da, ax, **kwargs):
    dim = get_ugrid_dim(grid, da)
    if dim != NODE:
        raise ValueError("tripcolor only supports data on nodes")
    (x, y, triangles), _ = grid.triangulation
    primitive = ax.tripcolor(x, y, triangles, da.to_numpy().ravel(), **kwargs)
    return primitive


@_plot2d
def line(grid, da, ax, **kwargs):
    from matplotlib.collections import LineCollection

    if da is not None:
        dim = get_ugrid_dim(grid, da)
        if dim != EDGE:
            raise ValueError("line only supports data on edges")
    else:
        dim = None

    edge_nodes = grid.edge_node_connectivity
    n_edge = len(edge_nodes)
    edge_coords = np.empty((n_edge, 2, 2), dtype=FloatDType)
    node_0 = edge_nodes[:, 0]
    node_1 = edge_nodes[:, 1]
    edge_coords[:, 0, 0] = grid.node_x[node_0]
    edge_coords[:, 0, 1] = grid.node_y[node_0]
    edge_coords[:, 1, 0] = grid.node_x[node_1]
    edge_coords[:, 1, 1] = grid.node_y[node_1]

    # PolyCollection takes a norm, but not vmin, vmax.
    norm = kwargs.get("norm", None)
    vmin = kwargs.pop("vmin", None)
    vmax = kwargs.pop("vmax", None)

    collection = LineCollection(edge_coords, **kwargs)

    if dim == EDGE:
        collection.set_array(da.to_numpy())
        collection._scale_norm(norm, vmin, vmax)

    primitive = ax.add_collection(collection, autolim=False)

    xmin, ymin, xmax, ymax = grid.bounds
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    return primitive


@_plot2d
def imshow(grid, da, ax, **kwargs):
    """
    Image plot of 2D DataArray.
    Wraps :py:func:`matplotlib:matplotlib.pyplot.imshow`.

    This rasterizes the grid before plotting. Pass a ``resolution`` keyword to
    control the rasterization resolution.
    """
    dim = get_ugrid_dim(grid, da)
    if dim != FACE:
        raise ValueError("imshow only supports data on faces")

    if "extent" not in kwargs:
        xmin, ymin, xmax, ymax = grid.bounds
        kwargs["extent"] = xmin, xmax, ymin, ymax
    else:
        if kwargs.get("origin", None) == "upper":
            xmin, xmax, ymin, ymax = kwargs["extent"]
        else:
            xmin, xmax, ymax, ymin = kwargs["extent"]

    dx = xmax - xmin
    dy = ymax - ymin

    # Check if a rasterization resolution is passed; Default to 500 raster
    # cells otherwise for the smallest axis.
    resolution = kwargs.pop("resolution", None)
    if resolution is None:
        resolution = min(dx, dy) / 500

    _, _, index = grid.rasterize(resolution)
    img = da.to_numpy()[index].astype(float)
    img[index == -1] = np.nan
    primitive = ax.imshow(img, **kwargs)
    return primitive


@_plot2d
def contour(grid, da, ax, **kwargs):
    """
    Create a contour plot of a 2D UgridDataArray.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.tricontour`.
    """
    dim = get_ugrid_dim(grid, da)
    if dim == NODE:
        (x, y, triangles), _ = grid.triangulation
        z = da
    elif dim == FACE:
        (x, y, triangles), index = grid.centroid_triangulation
        z = da.isel({grid.face_dimension: index})
    else:
        raise ValueError("contour only supports data on nodes or faces")

    primitive = ax.tricontour(x, y, triangles, z.to_numpy().ravel(), **kwargs)
    return primitive


@_plot2d
def contourf(grid, da, ax, **kwargs):
    """
    Create a filled contour plot of a 2D UgridDataArray.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.tricontourf`.
    """
    dim = get_ugrid_dim(grid, da)
    if dim == NODE:
        (x, y, triangles), _ = grid.triangulation
        z = da
    elif dim == FACE:
        (x, y, triangles), index = grid.centroid_triangulation
        z = da.isel({grid.face_dimension: index})
    else:
        raise ValueError("contourf only supports data on nodes or faces")

    primitive = ax.tricontourf(x, y, triangles, z.to_numpy().ravel(), **kwargs)
    return primitive


@_plot2d
def pcolormesh(grid, da, ax, **kwargs):
    """
    Create a pseudocolor mesh plot of a 2D UgridDataArray.

    Wraps matplotlib PolyCollection.
    """
    from matplotlib.collections import PolyCollection

    dim = get_ugrid_dim(grid, da)
    if dim != FACE:
        raise ValueError("pcolormesh only supports data on faces")

    nodes = grid.node_coordinates
    faces, _ = close_polygons(grid.face_node_connectivity)
    vertices = nodes[faces]

    # PolyCollection takes a norm, but not vmin, vmax.
    norm = kwargs.get("norm", None)
    vmin = kwargs.pop("vmin", None)
    vmax = kwargs.pop("vmax", None)
    kwargs["closed"] = False  # Vertices are closing.
    # Set edgecolor to face by default, since we get a lot of white lines,
    # especially for larger grids.
    if "edgecolors" not in kwargs:
        kwargs["edgecolors"] = "face"

    collection = PolyCollection(vertices, **kwargs)
    collection.set_array(da.to_numpy().ravel())
    collection._scale_norm(norm, vmin, vmax)
    primitive = ax.add_collection(collection, autolim=False)

    xmin, ymin, xmax, ymax = grid.bounds
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    corners = (xmin, ymin), (xmax, ymax)
    ax.update_datalim(corners)

    return primitive


@_plot2d
def surface(grid, da, ax, **kwargs):
    """
    Create a surface plot of a 2D UgridDataArray.

    Wraps :py:func:`matplotlib:mplot3d:plot_trisurf`.
    """
    dim = get_ugrid_dim(grid, da)
    if dim == NODE:
        (x, y, triangles), _ = grid.triangulation
        z = da
    elif dim == FACE:
        (x, y, triangles), index = grid.centroid_triangulation
        z = da.isel({grid.face_dimension: index})
    else:
        raise ValueError("surface only supports data on nodes or faces")

    primitive = ax.plot_trisurf(x, y, triangles, z.to_numpy().ravel(), **kwargs)
    return primitive


def plot(
    grid,
    darray,
    ax=None,
    **kwargs,
):
    """
    Plot the data of the DataArray on the Ugrid topology.

    Default plot of DataArray using :py:mod:`matplotlib:matplotlib.pyplot`.

    Calls xarray plotting function based on the topology dimension of the data.

    =============== ===========================
    Dimension       Plotting function
    =============== ===========================
    Face            :py:func:`xugrid.plot.pcolormesh`
    Edge            :py:func:`xugrid.plot.line`
    Node            :py:func:`xugrid.plot.tripcolor`
    =============== ===========================

    Parameters
    ----------
    darray : DataArray
    grid: Union[Ugrid1d, Ugrid2d]
    **kwargs : optional
        Additional keyword arguments for Matplotlib.
    """

    dim = darray.dims[0]
    kwargs["ax"] = ax
    if grid.topology_dimension == 1:
        if dim == grid.edge_dimension:
            return line(grid, darray, **kwargs)
        elif dim == grid.node_dimension:
            return scatter(grid, darray, **kwargs)
        else:
            raise ValueError("Data dimensions is not one of node or edge dimension.")
    elif grid.topology_dimension == 2:
        if dim == grid.face_dimension:
            return pcolormesh(grid, darray, **kwargs)
        elif dim == grid.node_dimension:
            return tripcolor(grid, darray, **kwargs)
        elif dim == grid.edge_dimension:
            return line(grid, darray, **kwargs)
        else:
            raise ValueError(
                "Data dimensions is not one of face, node, or edge dimension."
            )
    else:
        raise ValueError("Topology dimension is not 1 or 2")


class _PlotMethods:
    """
    Enables use of plot functions as attributes.
    For example UgridDataArray.ugrid.plot.pcolormesh()
    """

    __slots__ = ("grid", "darray")

    def __init__(self, obj):
        darray = obj.obj
        grid = obj.grid

        invalid = set(darray.dims) - grid.dims
        if invalid:
            raise ValueError(
                f"UgridDataArray contains non-topology dimensions: {invalid}.\n"
                f"Expected only one of {grid.dims}."
            )

        self.grid = grid
        self.darray = darray.squeeze().compute()

    def __call__(self, **kwargs):
        return plot(self.grid, self.darray, **kwargs)

    @functools.wraps(contour)
    def contour(self, *args, **kwargs):
        return contour(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(contourf)
    def contourf(self, *args, **kwargs):
        return contourf(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(imshow)
    def imshow(self, *args, **kwargs):
        return imshow(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(line)
    def line(self, *args, **kwargs):
        if self.darray.dims[0] == self.grid.edge_dimension:
            z = self.darray
        else:
            z = None
        return line(self.grid, z, *args, **kwargs)

    @functools.wraps(pcolormesh)
    def pcolormesh(self, *args, **kwargs):
        return pcolormesh(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(scatter)
    def scatter(self, *args, **kwargs):
        return scatter(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(surface)
    def surface(self, *args, **kwargs):
        return surface(self.grid, self.darray, *args, **kwargs)

    @functools.wraps(tripcolor)
    def tripcolor(self, *args, **kwargs):
        return tripcolor(self.grid, self.darray, *args, **kwargs)



from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import scipy.sparse
import xarray as xr

# from xugrid.plot.pyvista import to_pyvista_grid
from xugrid.core.accessorbase import AbstractUgridAccessor
from xugrid.core.utils import UncachedAccessor
from xugrid.core.wrap import UgridDataArray, UgridDataset
from xugrid.plot.plot import _PlotMethods
from xugrid.ugrid import connectivity
from xugrid.ugrid.interpolate import (
    interpolate_na_helper,
    laplace_interpolate,
)
from xugrid.ugrid.ugrid1d import Ugrid1d
from xugrid.ugrid.ugrid2d import Ugrid2d
from xugrid.ugrid.ugridbase import UgridType


class UgridDataArrayAccessor(AbstractUgridAccessor):
    """
    This "accessor" makes operations using the UGRID topology available via the
    ``.ugrid`` attribute for UgridDataArrays and UgridDatasets.
    """

    def __init__(self, obj: xr.DataArray, grid: UgridType):
        self.obj = obj
        self.grid = grid

    @property
    def grids(self) -> List[UgridType]:
        """
        The UGRID topology of this DataArry, as a list. Included for
        consistency with UgridDataset.
        """
        return [self.grid]

    @property
    def name(self) -> str:
        """Name of the UGRID topology of this DataArray."""
        return self.grid.name

    @property
    def names(self) -> List[str]:
        """
        Name of the UGRID topology, as a list. Included for consistency with
        UgridDataset.
        """
        return [self.grid.name]

    @property
    def topology(self) -> Dict[str, UgridType]:
        """Mapping from name to UGRID topology."""
        return {self.name: self.grid}

    @property
    def bounds(self) -> Dict[str, Tuple]:
        """
        Mapping from grid name to tuple containing ``minx, miny, maxx, maxy``
        values of the grid's node coordinates.
        """
        return {self.grid.name: self.grid.bounds}

    @property
    def total_bounds(self) -> Tuple:
        """
        Returns a tuple containing ``minx, miny, maxx, maxy`` values of the grid's
        node coordinates.
        """
        return next(iter(self.bounds.values()))

    plot = UncachedAccessor(_PlotMethods)

    def rename(self, name: str) -> UgridDataArray:
        """
        Give a new name to the UGRID topology and update the associated
        coordinate and dimension names in the DataArray.

        Parameters
        ----------
        name: str
            The new name of the topology.
        """
        obj = self.obj
        new_grid, name_dict = self.grid.rename(name, return_name_dict=True)
        to_rename = tuple(obj.coords) + tuple(obj.dims)
        new_obj = obj.rename({k: v for k, v in name_dict.items() if k in to_rename})
        return UgridDataArray(new_obj, new_grid)

    def assign_node_coords(self) -> UgridDataArray:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        return UgridDataArray(self.grid.assign_node_coords(self.obj), self.grid)

    def assign_edge_coords(self) -> UgridDataArray:
        """
        Assign edge coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        return UgridDataArray(self.grid.assign_edge_coords(self.obj), self.grid)

    def assign_face_coords(self) -> UgridDataArray:
        """
        Assign face coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        if self.grid.topology_dimension == 1:
            raise TypeError("Cannot set face coords from a Ugrid1D topology")
        return UgridDataArray(self.grid.assign_face_coords(self.obj), self.grid)

    def set_node_coords(self, node_x: str, node_y: str):
        """
        Given names of x and y coordinates of the nodes of an object, set them
        as the coordinates in the grid.

        Parameters
        ----------
        node_x: str
            Name of the x coordinate of the nodes in the object.
        node_y: str
            Name of the y coordinate of the nodes in the object.
        """
        self.grid.set_node_coords(node_x, node_y, self.obj)

    def sel(self, x=None, y=None):
        """
        Return a new object, a subselection in the UGRID x and y coordinates.

        The indexing for x and y always occurs orthogonally, i.e.:
        ``.sel(x=[0.0, 5.0], y=[10.0, 15.0])`` results in a four points. For
        vectorized indexing (equal to ``zip``ing through x and y), see
        ``.sel_points``.

        Depending on the nature of the x and y indexers, a xugrid or xarray
        object is returned:

        * slice without step: ``x=slice(-100, 100)``: returns xugrid object, a
          part of the unstructured grid.
        * slice with step: ``x=slice(-100, 100, 10)``: returns xarray object, a
          series of points (x=[-100, -90, -80, ..., 90, 100]).
        * a scalar: ``x=5.0``: returns xarray object, a point.
        * an array: ``x=[1.0, 15.0, 17.0]``: returns xarray object, a series of
          points.

        Parameters
        ----------
        x: float, 1d array, slice
        y: float, 1d array, slice

        Returns
        -------
        selection: Union[UgridDataArray, UgridDataset, xr.DataArray, xr.Dataset]
        """
        result = self.grid.sel(self.obj, x, y)
        if isinstance(result, tuple):
            return UgridDataArray(*result)
        else:
            return result

    def sel_points(
        self, x, y, method=None, out_of_bounds="warn", fill_value=np.nan, tolerance=None
    ):
        """
        Select points in the unstructured grid.

        For data on the core dimension (faces for Ugrid2d, edges for Ugrid1d),
        points are located by checking containment unless method is "nearest".
        For data on other dimension (e.g. nodes), the nearest entity is found.

        Parameters
        ----------
        x: 1d array of floats with shape ``(n_points,)``
        y: 1d array of floats with shape ``(n_points,)``
        obj: xr.DataArray or xr.Dataset
        method: str, {None, "nearest"}, optional
            * None: default, locate the core entity (face for Ugrid2d, edge for
              Ugrid1d) that contains each point. For secondary entities (nodes
              and edges for Ugrid2d; nodes for Ugrid1d), the nearest is always returned.
            * "nearest": locate the nearest entity to each point, including the core
              entity.

        out_of_bounds: str, default ``"warn"``
            What to do when points are located outside of any feature:

            * raise: raise a ValueError.
            * ignore: return ``fill_value`` for the out of bounds points.
            * warn: give a warning and return NaN for the out of bounds points.
            * drop: drop the out of bounds points. They may be identified
              via the ``index`` coordinate of the returned selection.
        fill_value: scalar, DataArray, Dataset, or callable, optional, default: np.nan
            Value to assign to out-of-bounds points if out_of_bounds is warn
            or ignore. Forwarded to xarray's ``.where()`` method.
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. This
            is a floating point precision criterion, thus cannot be directly be
            interpreted as a distance. If None, ``numba_celltree`` estimates an
            appropriate tolerance by multiplying the maximum diagonal of the
            bounding boxes with 1e-12.

        Returns
        -------
        selection: xr.DataArray or xr.Dataset
            The name of the topology is prefixed in the x, y coordinates
            and in a points dimension.
        """
        return self.grid.sel_points(
            self.obj, x, y, method, out_of_bounds, fill_value, tolerance
        )

    def rasterize(self, resolution: float) -> xr.DataArray:
        """
        Rasterize unstructured grid by sampling.

        Parameters
        ----------
        resolution: float
            Spacing in x and y.

        Returns
        -------
        rasterized: xr.DataArray
        """
        x, y, index = self.grid.rasterize(resolution)
        return self._raster(x, y, index)

    def rasterize_like(self, other: Union[xr.DataArray, xr.Dataset]) -> xr.DataArray:
        """
        Rasterize unstructured grid by sampling on the x and y coordinates
        of ``other``.

        Parameters
        ----------
        resolution: float
            Spacing in x and y.
        other: Union[xr.DataArray, xr.Dataset]
            Object to take x and y coordinates from.

        Returns
        -------
        rasterized: xr.DataArray
        """
        x, y, index = self.grid.rasterize_like(
            x=other["x"].to_numpy(),
            y=other["y"].to_numpy(),
        )
        return self._raster(x, y, index)

    def to_periodic(self):
        """
        Convert this grid to a periodic grid, where the rightmost boundary
        shares its nodes with the leftmost boundary.

        Returns
        -------
        periodic: UgridDataArray
        """
        grid, obj = self.grid.to_periodic(obj=self.obj)
        return UgridDataArray(obj, grid)

    def to_nonperiodic(self, xmax: float):
        """
        Convert this grid from a periodic grid (where the rightmost boundary shares its
        nodes with the leftmost boundary) to an aperiodic grid, where the leftmost nodes
        are separate from the rightmost nodes.

        Parameters
        ----------
        xmax: float
            The x-value of the newly created rightmost boundary nodes.

        Returns
        -------
        nonperiodic: UgridDataArray
        """
        grid, obj = self.grid.to_nonperiodic(xmax=xmax, obj=self.obj)
        return UgridDataArray(obj, grid)

    def _to_facet(self, facet: str, newdim: str):
        """
        Map the data from one facet to another.

        Parameters
        ----------
        facet: str
            node, edge, face
        newdim: str
            how to name the new dimension, e.g. three nodes
            per triangle face, two nodes per edge, etc.
        """
        grid = self.grid
        obj = self.obj

        gridfacets = grid.facets
        if facet not in gridfacets:
            raise ValueError(
                f"Cannot map to {facet} for a {type(grid).__name__} topology."
            )

        if newdim in obj.dims:
            raise ValueError(
                f"Dimension {newdim} already exists. Please provide a new dimension name."
            )

        source_dim = grid.dims.intersection(obj.dims).pop()
        target_dim = getattr(grid, f"{facet}_dimension")
        if source_dim == target_dim:
            raise ValueError(
                f"No conversion needed, data is already {facet}-associated."
            )

        # Find out on which facet we're currently located
        source = {v: k for k, v in gridfacets.items()}[source_dim]
        connectivity = grid.format_connectivity_as_dense(
            getattr(grid, f"{facet}_{source}_connectivity")
        )
        indexer = xr.DataArray(connectivity, dims=(target_dim, newdim))
        # Ensure the source dimension is not chunked for efficient indexing.
        obj = obj.chunk({source_dim: -1})
        # Set the fill values (-1) to NaN
        mapped = obj.isel({source_dim: indexer}).where(indexer != -1)
        return UgridDataArray(mapped, grid)

    def to_node(self, dim: str = "nmax"):
        """
        Map data to nodes.

        Creates a new dimension representing the contributing source elements
        for each node, as multiple faces/edges can connect to a single node.

        Parameters
        ----------
        dim : str, optional

        Returns
        -------
        mapped: UgridDataArray
            A new UgridDataArray with data mapped to the nodes of the grid.

        Examples
        --------
        Compute the mean elevation based on the surrounding faces for each node:

        >>> node_elevation = face_elevation.to_node().mean("contributors")
        """
        return self._to_facet("node", dim)

    def to_edge(self, dim: str = "nmax"):
        """
        Map data to edges.

        Creates a new dimension representing the contributing source elements
        for each node, as two nodes or two faces are connected to an edge.

        Parameters
        ----------
        dim : str, optional

        Returns
        -------
        mapped: UgridDataArray
            A new UgridDataArray with data mapped to the edges of the grid.

        Examples
        --------
        Compute the mean elevation based on the nodes for each edge:

        >>> edge_elevation = node_elevation.to_edge().mean("contributors")
        """
        return self._to_facet("edge", dim)

    def to_face(self, dim: str = "nmax"):
        """
        Map data to faces.

        Creates a new dimension representing the contributing source elements
        for each node, as two edges or multiple nodes are connected to a face.

        Parameters
        ----------
        dim : str, optional

        Returns
        -------
        mapped: UgridDataArray
            A new UgridDataArray with data mapped to the faces of the grid.

        Examples
        --------
        Compute the mean elevation based on the nodes for each face:

        >>> face_elevation = node_elevation.to_face().mean("contributors")
        """
        return self._to_facet("face", dim)

    def intersect_line(
        self, start: Sequence[float], end: Sequence[float]
    ) -> xr.DataArray:
        """
        Intersect a line with the grid of this data, and fetch the values of
        the intersected faces.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset
        start: sequence of two floats
            coordinate pair (x, y), designating the start point of the line.
        end: sequence of two floats
            coordinate pair (x, y), designating the end point of the line.

        Returns
        -------
        intersection: xr.DataArray
            The length along the line is returned as the "s" coordinate.
        """
        return self.grid.intersect_line(self.obj, start, end)

    def intersect_linestring(self, linestring) -> xr.DataArray:
        """
        Intersect the grid along a collection of linestrings. Returns a new DataArray
        with the values for each intersected segment.

        Parameters
        ----------
        linestring: shapely.LineString

        Returns
        -------
        intersection: xr.DataArray
            The length along the linestring is returned as the "s" coordinate.
        """
        return self.grid.intersect_linestring(self.obj, linestring)

    @property
    def crs(self):
        """
        The Coordinate Reference System (CRS) represented as a ``pyproj.CRS`` object.

        Returns None if the CRS is not set.

        Returns
        -------
        crs: dict
            A dictionary containing the name of the grid and its CRS.
        """
        return {self.grid.name: self.grid.crs}

    def set_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
        allow_override: bool = False,
    ):
        """
        Set the Coordinate Reference System (CRS) of a UGRID topology.

        NOTE: The underlying geometries are not transformed to this CRS. To
        transform the geometries to a new CRS, use the ``to_crs`` method.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted
            by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying the projection.
        allow_override : bool, default False
            If the the UGRID topology already has a CRS, allow to replace the
            existing CRS, even when both are not equal.
        """
        self.grid.set_crs(crs, epsg, allow_override)
        self.grid._update_coordinate_attrs(self.obj)

    def to_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
    ) -> UgridDataArray:
        """
        Transform geometries to a new coordinate reference system.
        Transform all geometries in an active geometry column to a different coordinate
        reference system. The ``crs`` attribute on the current Ugrid must
        be set. Either ``crs`` or ``epsg`` may be specified for output.

        This method will transform all points in all objects. It has no notion
        of projecting the cells. All segments joining points are assumed to be
        lines in the current projection, not geodesics. Objects crossing the
        dateline (or other projection boundary) will have undesirable behavior.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying output projection.

        Notes
        -----
        Node coordinates are always recomputed from the reprojected node positions.
        Face and edge coordinates, if present, are also recomputed from the new
        node positions. Coordinates not governed by the UGRID conventions
        are left untouched.
        """
        grid = self.grid.to_crs(crs, epsg)
        obj = grid._assign_derived_coords(self.obj)
        return UgridDataArray(obj, grid)

    def to_geodataframe(
        self, name: str = None, dim_order=None
    ) -> "geopandas.GeoDataFrame":  # type: ignore # noqa
        """
        Convert data and topology of one facet (node, edge, face) of the grid
        to a geopandas GeoDataFrame. This also determines the geometry type of
        the geodataframe:

        * node: point
        * edge: line
        * face: polygon

        Parameters
        ----------
        dim: str
            node, edge, or face dimension. Inferred for DataArray.
        name: str
            Name to give to the array (required if unnamed).
        dim_order:
            Hierarchical dimension order for the resulting dataframe. Array content is
            transposed to this order and then written out as flat vectors in contiguous
            order, so the last dimension in this list will be contiguous in the resulting
            DataFrame. This has a major influence on which operations are efficient on the
            resulting dataframe.

            If provided, must include all dimensions of this DataArray. By default,
            dimensions are sorted according to the DataArray dimensions order.

        Returns
        -------
        geodataframe: gpd.GeoDataFrame
        """
        import geopandas as gpd

        dim = self.obj.dims[-1]
        if name is not None:
            ds = self.obj.to_dataset(name=name)
        else:
            ds = self.obj.to_dataset()

        variables = [var for var in ds.data_vars if dim in ds.variables[var].dims]
        # TODO deal with time-dependent data, etc.
        # Basically requires checking which variables are static, which aren't.
        # For non-static, requires repeating all geometries.
        # Call reset_index on multi-index to generate them as regular columns.
        df = ds[variables].to_dataframe(dim_order=dim_order)
        geometry = self.grid.to_shapely(dim)
        return gpd.GeoDataFrame(df, geometry=geometry, crs=self.grid.crs)

    def reindex_like(
        self,
        other: Union[UgridType, UgridDataArray, UgridDataset],
        tolerance: float = 0.0,
    ):
        """
        Conform this object to match the topology of another object. The
        topologies must be exactly equivalent: only the order of the nodes,
        edges, and faces may differ.

        Dimension names must match for equivalent topologies.

        Parameters
        ----------
        other: Ugrid1d, Ugrid2d, UgridDataArray, UgridDataset
        obj: DataArray or Dataset
        tolerance: float, default value 0.0.
            Maximum distance between inexact coordinate matches.

        Returns
        -------
        reindexed: UgridDataArray
        """
        if isinstance(other, (Ugrid1d, Ugrid2d)):
            other_grid = other
        elif isinstance(other, (UgridDataArray, UgridDataset)):
            other_grid = other.ugrid.grid
        else:
            raise TypeError(
                "Expected Ugrid1d, Ugrid2d, UgridDataArray, or UgridDataset,"
                f"received instead: {type(other).__name__}"
            )
        new_obj = self.grid.reindex_like(other_grid, obj=self.obj, tolerance=tolerance)
        return UgridDataArray(new_obj, other_grid)

    def _binary_iterate(self, iterations: int, mask, value, border_value):
        if border_value == value:
            exterior = self.grid.exterior_faces
        else:
            exterior = None
        if mask is not None:
            mask = mask.to_numpy()

        obj = self.obj
        if isinstance(obj, xr.DataArray):
            output = connectivity._binary_iterate(
                self.grid.face_face_connectivity,
                obj.to_numpy(),
                value,
                iterations,
                mask,
                exterior,
                border_value,
            )
            da = obj.copy(data=output)
            return UgridDataArray(da, self.grid.copy())
        elif isinstance(obj, xr.Dataset):
            raise NotImplementedError
        else:
            raise ValueError("object should be a xr.DataArray")

    def binary_dilation(
        self,
        iterations: int = 1,
        mask=None,
        border_value=False,
    ):
        """
        Binary dilation can be used on a boolean array to expand the "shape" of
        features.

        Compare with :py:func:`scipy.ndimage.binary_dilation`.

        Parameters
        ----------
        iterations: int, default: 1
        mask: 1d array of bool, optional
        border_value: bool, default value: False

        Returns
        -------
        dilated: UgridDataArray
        """
        return self._binary_iterate(iterations, mask, True, border_value)

    def binary_erosion(
        self,
        iterations: int = 1,
        mask=None,
        border_value=False,
    ):
        """
        Binary erosion can be used on a boolean array to shrink the "shape" of
        features.

        Compare with :py:func:`scipy.ndimage.binary_erosion`.

        Parameters
        ----------
        iterations: int, default: 1
        mask: 1d array of bool, optional
        border_value: bool, default value: False

        Returns
        -------
        eroded: UgridDataArray
        """
        return self._binary_iterate(iterations, mask, False, border_value)

    def connected_components(self):
        """
        Every edge or face is given a component number. If all are connected,
        all will have the same number.

        Wraps :py:func:`scipy.sparse.csgraph.connected_components``.

        Returns
        -------
        labelled: UgridDataArray
        """
        _, labels = scipy.sparse.csgraph.connected_components(
            self.grid.face_face_connectivity
        )
        return UgridDataArray(
            xr.DataArray(labels, dims=[self.grid.face_dimension]),
            self.grid,
        )

    def reverse_cuthill_mckee(self):
        """
        Reduces bandwith of the connectivity matrix.

        Wraps :py:func:`scipy.sparse.csgraph.reverse_cuthill_mckee`.

        Returns
        -------
        reordered: Union[UgridDataArray, UgridDataset]
        """
        grid = self.grid
        reordered_grid, reordering = self.grid.reverse_cuthill_mckee()
        reordered_data = self.obj.isel({grid.face_dimension: reordering})
        # TODO: this might not work properly if e.g. centroids are stored in obj.
        # Not all metadata would be reordered.
        return UgridDataArray(
            reordered_data,
            reordered_grid,
        )

    def label_partitions(
        self,
        n_part: int,
    ):
        """
        Generate partition labels for this grid topology using METIS:
        https://github.com/KarypisLab/METIS

        This method utilizes the pymetis Python bindings:
        https://github.com/inducer/pymetis

        The data of the UgridDataArray are treated as weights.

        Parameters
        ----------
        n_part: integer
            The number of parts to partition the mesh.

        Returns
        -------
        partition_labels: UgridDataArray of integers
        """
        obj = self.obj
        grid = self.grid
        if obj.dims != (grid.core_dimension,):
            raise ValueError(
                "Weights must be associated with the core-dimension "
                f"of the grid: {grid.core_dimension}"
            )
        return grid.label_partitions(n_part=n_part, weights=obj.to_numpy())

    def interpolate_na(
        self,
        method: str = "nearest",
        max_distance: Optional[float] = None,
    ):
        """
        Fill in NaNs by interpolating.

        This function automatically finds the UGRID dimension and broadcasts
        over the other dimensions.

        Parameters
        ----------
        method: str, default is "nearest"
            Currently the only supported method.
        max_distance: nonnegative float, optional.
            Use ``None`` for no maximum distance.

        Returns
        -------
        filled: UgridDataArray of floats
        """

        if method != "nearest":
            raise ValueError(f'"{method}" is not a valid interpolator.')

        if max_distance is None:
            max_distance = np.inf

        grid = self.grid
        da = self.obj
        ugrid_dim = grid.find_ugrid_dim(da)

        da_filled = interpolate_na_helper(
            da,
            ugrid_dim=ugrid_dim,
            func=grid._nearest_interpolate,
            kwargs={
                "ugrid_dim": ugrid_dim,
                "max_distance": max_distance,
            },
        )
        return UgridDataArray(da_filled, grid)

    def laplace_interpolate(
        self,
        xy_weights: bool = True,
        direct_solve: bool = False,
        delta=0.0,
        relax=0.0,
        rtol=0.0,
        atol=1e-4,
        maxiter: int = 500,
    ):
        """
        Fill in NaNs by using Laplace interpolation.

        This function automatically finds the UGRID dimension and broadcasts
        over the other dimensions.

        This solves Laplace's equation where where there is no data, with data
        values functioning as fixed potential boundary conditions.

        Note that an iterative solver method will be required for large grids.
        In this case, some experimentation with the solver settings may be
        required to find a converging solution of sufficient accuracy. Refer to
        the documentation of :py:func:`scipy.sparse.linalg.spilu` and
        :py:func:`scipy.sparse.linalg.cg`.

        Data can be interpolated from nodes or faces. Direct interpolation of edge
        associated data is not allowed. Instead, create node associated data first,
        then translate that data to the edges.

        Parameters
        ----------
        xy_weights: bool, default False.
            Wether to use the inverse of the centroid to centroid distance in
            the coefficient matrix. If ``False``, defaults to uniform
            coefficients of 1 so that each face connection has equal weight.
        direct_solve: bool, optional, default ``False``
            Whether to use a direct or an iterative solver or a conjugate gradient
            solver. Direct method provides an exact answer, but are unsuitable
            for large problems.
        delta: float, default 0.0.
            ILU0 preconditioner non-diagonally dominant correction.
        relax: float, default 0.0.
            Modified ILU0 preconditioner relaxation factor.
        rtol: float, optional, default 0.0.
            Convergence tolerance for ``scipy.sparse.linalg.cg``.
        atol: float, optional, default 1.0e-4.
            Convergence tolerance for ``scipy.sparse.linalg.cg``.
        maxiter: int, default 500.
            Maximum number of iterations for ``scipy.sparse.linalg.cg``.

        Returns
        -------
        filled: UgridDataArray of floats
        """
        grid = self.grid
        da = self.obj

        grid = self.grid
        da = self.obj
        ugrid_dim = grid.find_ugrid_dim(da)
        if ugrid_dim == grid.edge_dimension:
            raise ValueError("Laplace interpolation along edges is not allowed.")

        connectivity = grid.get_connectivity_matrix(ugrid_dim, xy_weights=xy_weights)
        _, components_labels = scipy.sparse.csgraph.connected_components(connectivity)
        da_filled = interpolate_na_helper(
            da,
            ugrid_dim,
            func=laplace_interpolate,
            kwargs={
                "connectivity": connectivity,
                "use_weights": xy_weights,
                "components_labels": components_labels,
                "direct_solve": direct_solve,
                "delta": delta,
                "relax": relax,
                "rtol": rtol,
                "atol": atol,
                "maxiter": maxiter,
            },
        )
        return UgridDataArray(da_filled, grid)

    def to_dataset(self, optional_attributes: bool = False):
        """
        Convert this UgridDataArray or UgridDataset into a standard
        xarray.Dataset.

        The UGRID topology information is added as standard data variables.

        Parameters
        ----------
        optional_attributes: bool, default: False.
            Whether to generate the UGRID optional attributes.

        Returns
        -------
        dataset: UgridDataset
        """
        return self.grid.to_dataset(self.obj, optional_attributes)



from typing import Optional

import numba as nb
import numpy as np
import xarray as xr

import xugrid as xu
from xugrid.constants import FloatDType
from xugrid.ugrid import voronoi
from xugrid.ugrid.ugrid2d import Ugrid2d


@nb.njit(cache=True)
def replace_interpolated_weights(
    vertices,
    faces,
    face_index,
    weights,
    node_to_node_map,
    node_index_threshold,
):
    n, m = weights.shape
    for i in range(n):
        face = faces[face_index[i]]
        weights_row = weights[i]
        for j in range(m):
            p = face[j]
            w = weights_row[j]
            if (p < node_index_threshold) or (w <= 0):
                continue
            # Find the two surrounding nodes (q and r)
            index = p - node_index_threshold
            q, r = node_to_node_map[index]
            px, py = vertices[p]
            qx, qy = vertices[q]
            rx, ry = vertices[r]
            # Compute the euclidian distance to both
            p_q = np.sqrt((qx - px) ** 2 + (qy - py) ** 2)
            p_r = np.sqrt((rx - px) ** 2 + (ry - py) ** 2)
            total = p_q + p_r
            # Redistribute weight according to inverse distance.
            weight_q = (p_r / total) * w
            weight_r = (p_q / total) * w
            # Set weights to zero for p, and add to r and q weights.
            weights[i, j] = 0.0
            # Search for p and q
            for jj in range(m):
                node = face[jj]
                if node == q:
                    weights[i, jj] += weight_q
                if node == r:
                    weights[i, jj] += weight_r
    return


class UnstructuredGrid2d:
    """
    Stores only the grid topology.

    e.g. face -> face

    Parameters
    ----------
    grid: Ugrid2d
    """

    def __init__(self, obj):
        # TODO: do not omit type check on grid!
        if isinstance(obj, (xu.UgridDataArray, xu.UgridDataset)):
            self.ugrid_topology = obj.grid
        elif isinstance(obj, Ugrid2d):
            self.ugrid_topology = obj
        else:
            options = {"Ugrid2d", "UgridDataArray", "UgridDataset"}
            raise TypeError(
                f"Expected one of {options}, received: {type(obj).__name__}"
            )

    @property
    def ndim(self):
        return 1

    @property
    def dims(self):
        return (self.ugrid_topology.face_dimension,)

    @property
    def shape(self):
        return (self.ugrid_topology.n_face,)

    @property
    def size(self):
        return self.ugrid_topology.n_face

    @property
    def area(self):
        return self.ugrid_topology.area

    def convert_to(self, matched_type):
        if isinstance(self, matched_type):
            return self
        else:
            TypeError(f"Cannot convert UnstructuredGrid2d to {matched_type.__name__}")

    def overlap(self, other, relative: bool):
        """
        Parameters
        ----------
        other: UnstructuredGrid2d
        relative: bool
            Whether to divide by the original area. Used for e.g.
            first-order-conservative methods.

        Returns
        -------
        source_index: 1d np.ndarray of int
        target_index: 1d np.ndarray of int
        weights: 1d np.ndarray of float
        """
        (
            target_index,
            source_index,
            weights,
        ) = self.ugrid_topology.celltree.intersect_faces(
            vertices=other.ugrid_topology.node_coordinates,
            faces=other.ugrid_topology.face_node_connectivity,
            fill_value=other.ugrid_topology.fill_value,
        )
        if relative:
            weights /= self.area[source_index]
        return source_index, target_index, weights

    def locate_centroids(self, other, tolerance: Optional[float] = None):
        tree = self.ugrid_topology.celltree
        source_index = tree.locate_points(other.ugrid_topology.centroids, tolerance)
        inside = source_index != -1
        source_index = source_index[inside]
        target_index = np.arange(other.size, dtype=source_index.dtype)[inside]
        weight_values = np.ones_like(source_index, dtype=FloatDType)
        return source_index, target_index, weight_values

    def barycentric(self, other, tolerance: Optional[float] = None):
        points = other.ugrid_topology.centroids
        grid = self.ugrid_topology

        # Create a voronoi grid to get surrounding nodes as vertices
        (
            vertices,
            faces,
            node_to_face_index,
            node_to_node_map,
        ) = voronoi.voronoi_topology(
            grid.node_face_connectivity,
            grid.node_coordinates,
            grid.centroids,
            edge_face_connectivity=grid.edge_face_connectivity,
            edge_node_connectivity=grid.edge_node_connectivity,
            add_exterior=True,
            add_vertices=True,
            skip_concave=True,
        )

        voronoi_grid = Ugrid2d(
            vertices[:, 0],
            vertices[:, 1],
            -1,
            faces,
        )
        face_index, weights = voronoi_grid.compute_barycentric_weights(
            points, tolerance
        )

        # Find which nodes are interpolated. Redistribute their weights
        # according to distance to projection vertex.
        replace_interpolated_weights(
            vertices=vertices,
            faces=faces,
            face_index=face_index,
            weights=weights,
            node_to_node_map=node_to_node_map,
            node_index_threshold=len(vertices) - len(node_to_node_map),
        )

        # Discards 0 weights and points that fall outside of the grid.
        outside = grid.locate_points(points) == -1
        weights[outside] = 0
        keep = weights.ravel() > 0
        source_index = node_to_face_index[
            voronoi_grid.face_node_connectivity[face_index]
        ].ravel()[keep]

        n_points, n_max_node = weights.shape
        target_index = np.repeat(np.arange(n_points), n_max_node)[keep]
        weights = weights.ravel()[keep]

        order = np.argsort(target_index)
        return source_index[order], target_index[order], weights[order]

    def intersection_length(self, other: "xu.regrid.network.Network1d", relative: bool):
        (
            target_index,
            source_index,
            intersections,
        ) = self.ugrid_topology.celltree.intersect_edges(
            other.ugrid_topology.edge_node_coordinates
        )
        order = np.argsort(source_index)
        length = np.linalg.norm(np.diff(intersections, axis=1)[:, 0, :], axis=-1)
        if relative:
            length /= other.length[source_index]
        return target_index[order], source_index[order], length[order]

    def to_dataset(self, name: str):
        ds = self.ugrid_topology.rename(name).to_dataset()
        ds[name + "_type"] = xr.DataArray(-1, attrs={"type": "UnstructuredGrid2d"})
        return ds



import abc
import copy
import warnings
from itertools import chain
from typing import Any, Dict, Literal, Optional, Sequence, Set, Tuple, Type, Union, cast

import numpy as np
import pandas as pd
import xarray as xr
from numba_celltree import CellTree2d, EdgeCellTree2d
from numpy.typing import ArrayLike
from scipy.sparse import coo_matrix, csr_matrix
from scipy.spatial import KDTree

from xugrid.constants import FILL_VALUE, BoolArray, FloatArray, IntArray
from xugrid.ugrid import connectivity, conventions
from xugrid.ugrid.crs import CrsPlaceholder, crs_from_attrs, crs_to_attrs
from xugrid.ugrid.selection_utils import get_sorted_section_coords


def numeric_bound(v: Union[float, None], other: float):
    if v is None:
        return other
    else:
        return v


def as_pandas_index(index: Union[BoolArray, IntArray, pd.Index], n: int):
    if isinstance(index, np.ndarray):
        if index.size > n:
            raise ValueError(
                f"index size {index.size} is larger than dimension size: {n}"
            )
        if np.issubdtype(index.dtype, np.bool_):
            # Significantly quicker if all true.
            if index.all():
                pd_index = pd.RangeIndex(0, n)
            else:
                pd_index = pd.Index((np.arange(n)[index]))
        elif np.issubdtype(index.dtype, np.integer):
            pd_index = pd.Index(index)
        else:
            raise TypeError(f"index should be bool or integer. Received: {index.dtype}")

    elif isinstance(index, pd.Index):
        pd_index = index

    else:
        raise TypeError(
            "index should be pandas Index or numpy array. Received: "
            f"{type(index).__name__}"
        )

    if not pd_index.is_unique:
        raise ValueError(
            "index contains repeated values; only subsets will result "
            "in valid UGRID topology."
        )
    #    # TODO?
    #    # Uniqueness is required, but sorting arguably not.
    #    if not pd_index.is_monotonic_increasing:
    #        raise NotImplementedError("UGRID indexes must be sorted and unique.")

    return pd_index


def align(obj, grids, old_indexes):
    """
    Check which indexes have changed. Index on those new values.
    If none are changed, return (obj, grids) as is.
    """
    if old_indexes is None:
        return obj, grids

    ugrid_dims = set(chain.from_iterable(grid.dims for grid in grids)).intersection(
        old_indexes
    )
    new_indexes = {
        k: index
        for k, index in obj.indexes.items()
        if (k in ugrid_dims) and (not index.equals(old_indexes[k]))
    }
    if not new_indexes:
        return obj, grids

    # Group the indexers by grid
    new_grids = []
    for grid in grids:
        ugrid_dims = grid.dims.intersection(new_indexes)
        if ugrid_dims:
            ugrid_indexes = {dim: new_indexes[dim] for dim in ugrid_dims}
            newgrid, indexers = grid.isel(indexers=ugrid_indexes, return_index=True)
            indexers = {
                k: v
                for k, v in indexers.items()
                if k in obj.dims and k not in new_indexes
            }
            obj = obj.isel(indexers)
            new_grids.append(newgrid)
        else:
            # In case of multiple topologies, not every grid might be indexed.
            # In that case, just keep the untouched grids.
            new_grids.append(grid)

    return obj, new_grids


class AbstractUgrid(abc.ABC):
    @property
    @abc.abstractmethod
    def topology_dimension(self):
        pass

    @property
    @abc.abstractmethod
    def core_dimension(self):
        pass

    @property
    @abc.abstractmethod
    def dims(self) -> Set[str]:
        pass

    @property
    def dimensions(self) -> Dict[str, Dict[str, str]]:
        """
        Mapping from UGRID dimension names to lengths.

        This property will be changed to return a type more consistent with
        DataArray.dims in the future, i.e. a set of dimension names.
        """

        warnings.warn(
            ".dimensions will be removed and is replaced by .dims."
            "The return type of .dims is a set of dimension names. "
            "To access a mapping of names to lengths, use .sizes instead.",
            FutureWarning,
        )
        return self.sizes

    @property
    @abc.abstractmethod
    def sizes(self):
        pass

    @property
    @abc.abstractmethod
    def mesh(self):
        pass

    @property
    @abc.abstractmethod
    def meshkernel(self):
        pass

    @staticmethod
    @abc.abstractmethod
    def from_dataset(self):
        pass

    @abc.abstractmethod
    def to_dataset(self) -> xr.Dataset:
        pass

    @abc.abstractmethod
    def topology_subset(self):
        pass

    @abc.abstractmethod
    def clip_box(self):
        pass

    @abc.abstractmethod
    def _clear_geometry_properties(self):
        pass

    @abc.abstractmethod
    def _assign_derived_coords(self, obj):
        pass

    @staticmethod
    @abc.abstractmethod
    def merge_partitions():
        pass

    @abc.abstractmethod
    def reindex_like(self):
        pass

    @abc.abstractmethod
    def get_connectivity_matrix(self, dim: str, xy_weights: bool) -> csr_matrix:
        pass

    @abc.abstractmethod
    def create_data_array(self, data: ArrayLike, facet: str):
        pass

    @abc.abstractmethod
    def get_coordinates(self, dim: str):
        pass

    @staticmethod
    @abc.abstractmethod
    def _section_coordinates(
        edges: FloatArray, xy: FloatArray, dim: str, index: IntArray, name: str
    ):
        pass

    @staticmethod
    def format_connectivity_as_dense(
        sparse_connectivity: Union[csr_matrix, coo_matrix, np.ndarray],
    ):
        """
        Return a rectangular representation of a sparse (CSR or COO)
        connectivity array. The number of columns in the dense array will be
        determined by the maximum number of connections for any row in the
        sparse matrix. Irregular connectivities are represented by a fill value
        of -1.

        Parameters
        ----------
        sparse_connectivity: scipy.sparse.csr_matrix, scipy.sparse.coo_matrix, np.ndarray
            Connectivity in CSR or COO form. Connectivities already in dense
            rectangular form are returned as-is.

        Returns
        -------
        dense: np.ndarray
            Connectivity in rectangular form.
        """
        if isinstance(sparse_connectivity, np.ndarray):
            return sparse_connectivity
        else:
            return connectivity.to_dense(sparse_connectivity)

    @staticmethod
    def format_connectivity_as_sparse(dense_connectivity: np.ndarray) -> csr_matrix:
        """
        Convert a rectangular dense connectivity array to a sparse CSR matrix.

        This function converts a dense rectangular array with fill values
        representing irregular connectivities into a sparse CSR matrix where
        only the valid connections are stored. Irregular connectivities must be
        marked with a fill value of -1.

        Parameters
        ----------
        dense_connectivity: np.ndarray
            Connectivity in rectangular dense form, where each row contains the
            connections for an entity and -1 represents missing connections.

        Returns
        -------
        sparse: scipy.sparse.csr_matrix
            Connectivity in CSR sparse format.
        """
        if isinstance(dense_connectivity, csr_matrix):
            return dense_connectivity
        elif isinstance(dense_connectivity, coo_matrix):
            return dense_connectivity.tocsr()
        else:
            return connectivity.to_sparse(dense_connectivity)

    def _create_data_array(self, data: ArrayLike, dimension: str):
        from xugrid import UgridDataArray

        data = np.array(data)
        if data.ndim != 1:
            raise ValueError(
                "Can only create DataArrays from 1D arrays. "
                f"Data has {data.ndim} dimensions."
            )
        len_data = len(data)
        len_grid = self.sizes[dimension]
        if len_data != len_grid:
            raise ValueError(
                f"Conflicting sizes for dimension {dimension}: length "
                f"{len_data} on the data, but length {len_grid} on the grid."
            )

        da = xr.DataArray(data=data, dims=(dimension,))

        # TODO: is there a better way to do this to satisfy mypy?
        grid = cast(UgridType, self)
        return UgridDataArray(da, grid)

    def _initialize_indexes_attrs(self, name, dataset, indexes, attrs):
        defaults = conventions.default_topology_attrs(name, self.topology_dimension)

        if dataset is None:
            if attrs is None:
                x, y = defaults["node_coordinates"].split()
                indexes = {"node_x": x, "node_y": y}
            else:
                if indexes is None:
                    raise ValueError("indexes must be provided for attrs")
                defaults.update(attrs)

            self._indexes = indexes
            self._attrs = defaults

        else:
            if attrs is not None:
                raise ValueError("Provide either dataset or attrs, not both.")
            if indexes is None:
                raise ValueError("indexes must be provided for dataset")

            derived_dims = dataset.ugrid_roles.dimensions[name]
            self._indexes = indexes
            self._attrs = {**defaults, **derived_dims, **dataset[name].attrs}

        # Ensure the name is always in sync.
        self._attrs["name"] = name
        return

    def rename(self, name: str, return_name_dict: bool = False):
        """
        Create a new grid with all variables named according to the default
        naming conventions.
        """
        # Get the old and the new names. Their keys are the same.
        old_attrs = self._attrs
        new_attrs = conventions.default_topology_attrs(name, self.topology_dimension)

        # The attrs will have some roles joined together, e.g. node_coordinates
        # will contain x and y as "mesh2d_node_x mesh2d_node_y".
        name_dict = {self.name: name}
        skip = ("cf_role", "long_name", "topology_dimension")
        for key, value in old_attrs.items():
            if key in new_attrs and key not in skip:
                split_new = new_attrs[key].split()
                split_old = value.split()
                if len(split_new) != len(split_old):
                    raise ValueError(
                        f"Number of entries does not match on {key}: "
                        f"{split_new} versus {split_old}"
                    )
                for name_key, name_value in zip(split_old, split_new):
                    name_dict[name_key] = name_value

        new = self.copy()
        new.name = name
        new._attrs = new_attrs
        new._indexes = {k: name_dict[v] for k, v in new._indexes.items()}
        if new._dataset is not None:
            to_rename = (
                tuple(new._dataset.data_vars)
                + tuple(new._dataset.coords)
                + tuple(new._dataset.dims)
            )
            new._dataset = new._dataset.rename(
                {k: v for k, v in name_dict.items() if k in to_rename}
            )

        if return_name_dict:
            return new, name_dict
        else:
            return new

    def _propagate_properties(self, other) -> None:
        other.start_index = self.start_index
        other.fill_value = self.fill_value

    @staticmethod
    def _single_topology(dataset: xr.Dataset):
        topologies = dataset.ugrid_roles.topology
        n_topology = len(topologies)
        if n_topology == 0:
            raise ValueError("Dataset contains no UGRID topology variable.")
        elif n_topology > 1:
            raise ValueError(
                f"Dataset contains {n_topology} topology variables, "
                "please specify the topology variable name to use."
            )
        return topologies[0]

    def _filtered_attrs(self, dataset: xr.Dataset):
        """Remove names that are not present in the dataset."""
        topodim = self.topology_dimension
        attrs = self._attrs.copy()

        ugrid_dims = conventions._DIM_NAMES[topodim] + tuple(
            [dims[0] for dims in conventions._CONNECTIVITY_DIMS.values()]
        )
        for key in ugrid_dims:
            if key in attrs and attrs[key] not in dataset.dims:
                attrs.pop(key)

        for key in conventions._CONNECTIVITY_NAMES[topodim]:
            if key in attrs and attrs[key] not in dataset:
                attrs.pop(key)

        for coord in conventions._COORD_NAMES[topodim]:
            if coord in attrs:
                names = attrs[coord].split(" ")
                present = [name for name in names if name in dataset]
                if present:
                    attrs[coord] = " ".join(present)
                else:
                    attrs.pop(coord)

        return attrs

    @staticmethod
    def _extract_crs(dataset: xr.Dataset, topology: str):
        grid_mapping_name = dataset.ugrid_roles.grid_mapping_names[topology]
        stdname_projected = dataset.ugrid_roles.is_projected[topology]
        if grid_mapping_name is not None:
            crs = crs_from_attrs(dataset[grid_mapping_name].attrs)
        else:
            crs = None

        if not (crs is None or isinstance(crs, CrsPlaceholder)):
            # Then it's pyproj
            is_projected = crs.is_projected
            if stdname_projected is not None and stdname_projected != is_projected:
                warnings.warn(
                    f"standard_name suggests {'projected' if stdname_projected else 'geographic'} "
                    f"coordinates, but the CRS ({crs}) is "
                    f"{'projected' if is_projected else 'geographic'}. "
                    "The CRS will take priority.",
                    UserWarning,
                    stacklevel=2,
                )
            return crs, is_projected

        if stdname_projected is not None:
            is_projected = stdname_projected
        else:
            warnings.warn(
                f"No CRS or recognizable standard_name found for topology '{topology}'. "
                "Assuming projected coordinates.",
                UserWarning,
                stacklevel=2,
            )
            is_projected = True
        return crs, is_projected

    @staticmethod
    def _validate_crs(crs: Any, is_projected: bool):
        if crs is None or isinstance(crs, CrsPlaceholder):
            _crs = crs
            _is_projected = is_projected
        else:
            import pyproj

            _crs = pyproj.CRS.from_user_input(crs)
            if not (_crs.is_projected ^ _crs.is_geographic):
                raise ValueError(
                    f"Unsupported CRS: {crs}.\n"
                    "CRS should either be geographic (latitude / longitude) or projected.\n"
                    "If your CRS is neither, please file an issue at https://github.com/deltares/xugrid/issues"
                )
            _is_projected = _crs.is_projected
        return _crs, _is_projected

    def write_grid_mapping(self, dataset, grid_mapping_name: str = None):
        """
        Use the CRS of the topology, if set, to write the CF grid_mapping
        attributes to a newly created mapping variable.

        The grid mapping variable is stamped onto all data variables that share
        dimensions with this topology via the ``grid_mapping`` attribute, following
        CF conventions. Additionally, coordinate variables are also stamped, which
        is not strictly CF-compliant but is required for compatibility for e.g.
        QGIS-MDAL, which reads ``grid_mapping`` from coordinates rather than
        data variables.

        Parameters
        ----------
        dataset: xr.Dataset
        grid_mapping_name: str, optional
            Name of the CF grid mapping variable. If not provided (or None),
            defaults to ``f"{self.name}_crs"`` for this topology.

        Returns
        -------
        dataset: xr.Dataset
            Modified dataset with CF compliant CRS information.
        """

        if self.crs is not None:
            dataset = dataset.copy()
            if grid_mapping_name is None:
                grid_mapping_name = f"{self.name}_crs"

            # Needs to be this int value for DFM/Interacter...
            # TODO: might change if DFM is changed; we could then simply use 0.
            _FILL_ = np.int32(np.iinfo(np.int32).min + 1)
            dataset[grid_mapping_name] = xr.Variable(
                (), _FILL_, attrs=crs_to_attrs(self.crs)
            )
            # Stamp grid_mapping on data variables that sit on this topology
            # QGIS-MDAL reads the grid_mapping off the coordinates, hence
            # we will stamp those as well (.variables instead of .data_vars).
            for var in dataset.variables.values():
                if set(self.dims) & set(var.dims):
                    var.attrs["grid_mapping"] = grid_mapping_name
        return dataset

    def __repr__(self):
        if self._dataset:
            return self._dataset.__repr__()
        else:
            return self.to_dataset().__repr__()

    def equals(self, other) -> bool:
        if other is self:
            return True
        elif isinstance(other, type(self)):
            xr_self = self.to_dataset()
            xr_other = other.to_dataset()
            return xr_self.identical(xr_other)
        return False

    def copy(self):
        """Create a deepcopy."""
        return copy.deepcopy(self)

    @property
    def fill_value(self) -> int:
        """Fill value for UGRID connectivity arrays."""
        return self._fill_value

    @fill_value.setter
    def fill_value(self, value: int):
        self._fill_value = value

    @property
    def start_index(self) -> int:
        """Start index for UGRID connectivity arrays."""
        return self._start_index

    @start_index.setter
    def start_index(self, value: Literal[0, 1]):
        if value not in (0, 1):
            raise ValueError(f"start_index must be 0 or 1, received: {value}")
        self._start_index = value

    @property
    def attrs(self):
        return copy.deepcopy(self._attrs)

    @property
    def node_dimension(self):
        """Name of node dimension"""
        return self._attrs["node_dimension"]

    @property
    def edge_dimension(self):
        """Name of edge dimension"""
        return self._attrs["edge_dimension"]

    @property
    def max_connectivity_dimensions(self) -> tuple[str]:
        return ()

    @property
    def max_connectivity_sizes(self) -> dict[str, int]:
        return {}

    @property
    def node_coordinates(self) -> FloatArray:
        """Coordinates (x, y) of the nodes (vertices)"""
        return np.column_stack([self.node_x, self.node_y])

    @property
    def n_node(self) -> int:
        """Number of nodes (vertices) in the UGRID topology"""
        return self.node_x.size

    @property
    def n_edge(self) -> int:
        """Number of edges in the UGRID topology"""
        return self.edge_node_connectivity.shape[0]

    @property
    def edge_x(self):
        """x-coordinate of every edge in the UGRID topology"""
        if self._edge_x is None:
            self._edge_x = self.node_x[self.edge_node_connectivity].mean(axis=1)
        return self._edge_x

    @property
    def edge_y(self):
        """y-coordinate of every edge in the UGRID topology"""
        if self._edge_y is None:
            self._edge_y = self.node_y[self.edge_node_connectivity].mean(axis=1)
        return self._edge_y

    @property
    def edge_coordinates(self) -> FloatArray:
        """Centroid (x,y) coordinates of every edge in the UGRID topology"""
        return np.column_stack([self.edge_x, self.edge_y])

    @property
    def edge_node_coordinates(self) -> FloatArray:
        """Node coordinates for every edge, shape: ``n_edge, 2, 2``."""
        return self.node_coordinates[self.edge_node_connectivity]

    @property
    @abc.abstractmethod
    def coords(self):
        pass

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        """Returns a tuple with the node bounds: xmin, ymin, xmax, ymax"""
        if any(
            [
                self._xmin is None,
                self._ymin is None,
                self._xmax is None,
                self._ymax is None,
            ]
        ):
            self._xmin = self.node_x.min()
            self._ymin = self.node_y.min()
            self._xmax = self.node_x.max()
            self._ymax = self.node_y.max()
        return (
            self._xmin,
            self._ymin,
            self._xmax,
            self._ymax,
        )

    @property
    def edge_bounds(self) -> FloatArray:
        """
        Returns a numpy array with columns ``minx, miny, maxx, maxy``,
        describing the bounds of every edge in the grid.

        Returns
        -------
        edge_bounds: np.ndarray of shape (n_edge, 4)
        """
        x = self.node_x[self.edge_node_connectivity]
        y = self.node_y[self.edge_node_connectivity]
        return np.column_stack(
            [
                x.min(axis=1),
                y.min(axis=1),
                x.max(axis=1),
                y.max(axis=1),
            ]
        )

    @staticmethod
    def _prepare_connectivity(
        da: xr.DataArray, fill_value: Union[int, float], dtype: type
    ) -> xr.DataArray:
        """
        Undo the work xarray does when it encounters a _FillValue for UGRID
        connectivity arrays. Set an external unified value back (across all
        connectivities!), and cast back to the desired dtype.
        """
        data = da.to_numpy().copy()
        # If xarray detects a _FillValue, it converts the array to floats and
        # replaces the fill value by NaN, and moves the _FillValue to
        # da.encoding.
        if "_FillValue" in da.attrs:
            is_fill = data == da.attrs["_FillValue"]
        else:
            is_fill = np.isnan(data)
        # Set the fill_value before casting: otherwise the cast may fail.
        data[is_fill] = fill_value
        cast = data.astype(dtype, copy=False)
        not_fill = ~is_fill
        if (cast[not_fill] < 0).any():
            raise ValueError("connectivity contains negative values")
        return da.copy(data=cast)

    def _adjust_connectivity(self, connectivity: IntArray) -> IntArray:
        """Adjust connectivity for desired fill_value and start_index."""
        c = connectivity.copy()
        if self.start_index == 0 and self.fill_value == FILL_VALUE:
            return c
        is_fill = c == FILL_VALUE
        if self.start_index:
            c[~is_fill] += self.start_index
        if self.fill_value != FILL_VALUE:
            c[is_fill] = self.fill_value
        return c

    def _precheck(self, multi_index):
        dim, index = multi_index.popitem()
        for check_dim, check_index in multi_index.items():
            if not index.equals(check_index):
                raise ValueError(
                    f"UGRID dimensions do not align: {dim} versus {check_dim}"
                )
        return index

    def _postcheck(self, indexers, finalized_indexers):
        for dim, indexer in indexers.items():
            if dim != self.core_dimension:
                if not indexer.equals(finalized_indexers[dim]):
                    raise ValueError(
                        f"This subset selection of UGRID dimension {dim} results "
                        "in an invalid topology "
                    )
        return

    def find_ugrid_dim(self, obj: Union[xr.DataArray, xr.Dataset]):
        """Find the UGRID dimension that is present in the object."""
        ugrid_dims = self.dims.intersection(obj.dims)
        if len(ugrid_dims) != 1:
            raise ValueError(
                "UgridDataArray should contain exactly one of the UGRID "
                f"dimensions: {self.dims}"
            )
        return ugrid_dims.pop()

    def set_node_coords(
        self,
        node_x: str,
        node_y: str,
        obj: Union[xr.DataArray, xr.Dataset],
        is_projected: bool = True,
        crs: Any = None,
    ):
        """
        Given names of x and y coordinates of the nodes of an object, set them
        as the coordinates in the grid.

        Parameters
        ----------
        node_x: str
            Name of the x coordinate of the nodes in the object.
        node_y: str
            Name of the y coordinate of the nodes in the object.
        obj: xr.DataArray, xr.Dataset
        is_projected: bool, optional
            Whether node_x and node_y are longitude and latitude or projected x and
            y coordinates. Used to write the appropriate standard_name in the
            coordinate attributes. If crs is provided, its value will take priority.
        crs: Any, optional
            Coordinate Reference System of the geometry objects. Can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        """
        if " " in node_x or " " in node_y:
            raise ValueError("coordinate names may not contain spaces")

        x = obj[node_x].to_numpy()
        y = obj[node_y].to_numpy()

        if (x.ndim != 1) or (x.size != self.n_node):
            raise ValueError(
                "shape of node_x does not match n_node of grid: "
                f"{x.shape} versus {self.n_node}"
            )
        if (y.ndim != 1) or (y.size != self.n_node):
            raise ValueError(
                "shape of node_y does not match n_node of grid: "
                f"{y.shape} versus {self.n_node}"
            )

        # Remove them, then append at the end.
        node_coords = [
            coord
            for coord in self._attrs["node_coordinates"].split(" ")
            if coord not in (node_x, node_y)
        ]
        node_coords.extend((node_x, node_y))

        self._clear_geometry_properties()
        self.node_x = np.ascontiguousarray(x)
        self.node_y = np.ascontiguousarray(y)
        self._attrs["node_coordinates"] = " ".join(node_coords)
        self._indexes["node_x"] = node_x
        self._indexes["node_y"] = node_y
        self.crs, self.is_projected = self._validate_crs(crs, is_projected)

    def assign_node_coords(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset

        Returns
        -------
        assigned (same type as obj)
        """
        xname = self._indexes["node_x"]
        yname = self._indexes["node_y"]
        x_attrs = conventions.DEFAULT_ATTRS["node_x"][self.is_projected]
        y_attrs = conventions.DEFAULT_ATTRS["node_y"][self.is_projected]
        coords = {
            xname: xr.DataArray(
                data=self.node_x,
                dims=(self.node_dimension,),
                attrs=x_attrs,
            ),
            yname: xr.DataArray(
                data=self.node_y,
                dims=(self.node_dimension,),
                attrs=y_attrs,
            ),
        }
        return obj.assign_coords(coords)

    def assign_edge_coords(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset

        Returns
        -------
        assigned (same type as obj)
        """
        xname = self._indexes.get("edge_x", f"{self.name}_edge_x")
        yname = self._indexes.get("edge_y", f"{self.name}_edge_y")
        x_attrs = conventions.DEFAULT_ATTRS["edge_x"][self.is_projected]
        y_attrs = conventions.DEFAULT_ATTRS["edge_y"][self.is_projected]
        coords = {
            xname: xr.DataArray(
                data=self.edge_x,
                dims=(self.edge_dimension,),
                attrs=x_attrs,
            ),
            yname: xr.DataArray(
                data=self.edge_y,
                dims=(self.edge_dimension,),
                attrs=y_attrs,
            ),
        }
        return obj.assign_coords(coords)

    @property
    def node_edge_connectivity(self) -> csr_matrix:
        """
        Node to edge connectivity.

        Returns
        -------
        connectivity: csr_matrix
        """
        if self._node_edge_connectivity is None:
            self._node_edge_connectivity = connectivity.invert_dense_to_sparse(
                self.edge_node_connectivity
            )
        return self._node_edge_connectivity

    @property
    def node_node_connectivity(self) -> csr_matrix:
        """
        Node to node connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) node index. The data of
        the matrix contains the edge index as every connection is formed by an
        edge.

        Returns
        -------
        connectivity: csr_matrix
        """
        if self._node_node_connectivity is None:
            self._node_node_connectivity = connectivity.node_node_connectivity(
                self.edge_node_connectivity
            )
        return self._node_node_connectivity

    @property
    def edge_edge_connectivity(self) -> csr_matrix:
        """
        Edge to edge connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) edge index. The data of
        the matrix contains the node index of the common node through which
        the connection is formed.

        Returns
        -------
        connectivity: csr_matrix
        """
        return connectivity.edge_edge_connectivity(
            self.edge_node_connectivity,
            self.node_edge_connectivity,
        )

    @property
    def directed_node_node_connectivity(self) -> csr_matrix:
        """
        Directed node to node connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) node index. The data of
        the matrix contains the edge index as every connection is formed by an
        edge.

        Returns
        -------
        connectivity: csr_matrix
        """
        return connectivity.directed_node_node_connectivity(self.edge_node_connectivity)

    @property
    def directed_edge_edge_connectivity(self) -> csr_matrix:
        """
        Directed edge to edge connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) edge index. The data of
        the matrix contains the node index of the common node through which
        the connection is formed.

        Returns
        -------
        connectivity: csr_matrix
        """
        return connectivity.directed_edge_edge_connectivity(
            self.edge_node_connectivity,
            self.node_edge_connectivity,
        )

    @property
    def edge_length(self):
        """Length of each edge."""
        return np.linalg.norm(
            np.diff(self.edge_node_coordinates, axis=1)[:, 0, :], axis=-1
        )

    @staticmethod
    def _connectivity_weights(connectivity: csr_matrix, coordinates: FloatArray):
        xy = coordinates
        coo = connectivity.tocoo()
        i = coo.row
        j = coo.col
        distance = np.linalg.norm(xy[j] - xy[i], axis=1)
        # The inverse distance is a measure of the strength of the connection.
        # Normalize so the weights are around 1.0
        return distance.mean() / distance

    def _update_coordinate_attrs(self, obj: xr.DataArray | xr.Dataset):
        for role, name in self._indexes.items():
            if name in obj.coords:
                obj[name].attrs = conventions.DEFAULT_ATTRS[role][self.is_projected]
            # Also update internal storage where data is preserved for roundtripping.
            if self._dataset is not None and name in self._dataset.coords:
                self._dataset[name].attrs = conventions.DEFAULT_ATTRS[role][
                    self.is_projected
                ]
        return

    def set_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
        allow_override: bool = False,
    ):
        """
        Set the Coordinate Reference System (CRS) of a UGRID topology.

        NOTE: The underlying geometries are not transformed to this CRS. To
        transform the geometries to a new CRS, use the ``to_crs`` method.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted
            by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying the projection.
        allow_override : bool, default False
            If the the UGRID topology already has a CRS, allow to replace the
            existing CRS, even when both are not equal.
        """
        import pyproj

        if crs is not None:
            crs = pyproj.CRS.from_user_input(crs)
        elif epsg is not None:
            crs = pyproj.CRS.from_epsg(epsg)
        else:
            raise ValueError("Must pass either crs or epsg.")
        crs, is_projected = self._validate_crs(crs, crs.is_projected)

        if not allow_override and self.crs is not None and not self.crs == crs:
            raise ValueError(
                "The Ugrid already has a CRS which is not equal to the passed "
                "CRS. Specify 'allow_override=True' to allow replacing the existing "
                "CRS without doing any transformation. If you actually want to "
                "transform the geometries, use '.to_crs' instead."
            )
        self.crs = crs
        self.is_projected = is_projected

    def to_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
    ):
        """
        Transform geometries to a new coordinate reference system.
        Transform all geometries in an active geometry column to a different coordinate
        reference system. The ``crs`` attribute on the current Ugrid must
        be set. Either ``crs`` or ``epsg`` may be specified for output.

        This method will transform all points in all objects. It has no notion
        of projecting the cells. All segments joining points are assumed to be
        lines in the current projection, not geodesics. Objects crossing the
        dateline (or other projection boundary) will have undesirable behavior.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying output projection.
        """
        import pyproj

        if self.crs is None:
            raise ValueError(
                "Cannot transform naive geometries.  "
                "Please set a crs on the object first."
            )
        elif isinstance(self.crs, CrsPlaceholder):
            raise ValueError(
                "Cannot transform geometries: the current CRS is a placeholder and has "
                "not been parsed.\nThis may be because pyproj is not installed, or because pyproj "
                "failed to interpret the grid mapping attributes.\n"
                "Use .set_crs(..., allow_override=True) to set a valid pyproj CRS explicitly.\n"
            )

        if crs is not None:
            crs = pyproj.CRS.from_user_input(crs)
        elif epsg is not None:
            crs = pyproj.CRS.from_epsg(epsg)
        else:
            raise ValueError("Must pass either crs or epsg.")

        crs, is_projected = self._validate_crs(crs, crs.is_projected)
        grid = self.copy()
        if self.crs.is_exact_same(crs):
            return grid

        transformer = pyproj.Transformer.from_crs(
            crs_from=self.crs, crs_to=crs, always_xy=True
        )
        node_x, node_y = transformer.transform(xx=grid.node_x, yy=grid.node_y)
        grid.node_x = node_x
        grid.node_y = node_y
        grid._clear_geometry_properties()
        grid._dataset = None
        grid.crs = crs
        grid.is_projected = is_projected
        return grid

    @property
    def is_geographic(self):
        return not self.is_projected

    def plot(self, **kwargs):
        """
        Plot the edges of the mesh.

        Parameters
        ----------
        **kwargs : optional
            Additional keyword arguments to ``matplotlib.pyplot.line``.
        """
        from xugrid.plot import line

        return line(self, **kwargs)

    @property
    @abc.abstractmethod
    def celltree(self) -> Union[EdgeCellTree2d, CellTree2d]:
        raise NotImplementedError("Celltree must be implemented in subclass")

    @property
    def node_kdtree(self):
        if self._node_kdtree is None:
            self._node_kdtree = KDTree(self.node_coordinates)
        return self._node_kdtree

    @property
    def edge_kdtree(self):
        if self._edge_kdtree is None:
            self._edge_kdtree = KDTree(self.edge_coordinates)
        return self._edge_kdtree

    def sel_points(
        self,
        obj,
        x: FloatArray,
        y: FloatArray,
        method: str | None = None,
        out_of_bounds="warn",
        fill_value=np.nan,
        tolerance: float | None = None,
    ):
        """
        Select points in the unstructured grid.

        For data on the core dimension (faces for Ugrid2d, edges for Ugrid1d),
        points are located by checking containment unless method is "nearest".
        For data on other dimension (e.g. nodes), the nearest entity is found.

        Parameters
        ----------
        x: 1d array of floats with shape ``(n_points,)``
        y: 1d array of floats with shape ``(n_points,)``
        obj: xr.DataArray or xr.Dataset
        method: str, {None, "nearest"}, optional
            * None: default, locate the core entity (face for Ugrid2d, edge for
              Ugrid1d) that contains each point. For secondary entities (nodes
              and edges for Ugrid2d; nodes for Ugrid1d), the nearest is always returned.
            * "nearest": locate the nearest entity to each point, including the core
              entity.

        out_of_bounds: str, default ``"warn"``
            What to do when points are located outside of any feature:

            * raise: raise a ValueError.
            * ignore: return ``fill_value`` for the out of bounds points.
            * warn: give a warning and return NaN for the out of bounds points.
            * drop: drop the out of bounds points. They may be identified
              via the ``index`` coordinate of the returned selection.
        fill_value: scalar, DataArray, Dataset, or callable, optional, default: np.nan
            Value to assign to out-of-bounds points if out_of_bounds is warn
            or ignore. Forwarded to xarray's ``.where()`` method.
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. This
            is a floating point precision criterion, thus cannot be directly be
            interpreted as a distance. If None, ``numba_celltree`` estimates an
            appropriate tolerance by multiplying the maximum diagonal of the
            bounding boxes with 1e-12.

        Returns
        -------
        selection: xr.DataArray or xr.Dataset
            The name of the topology is prefixed in the x, y coordinates
            and in a points dimension.
        """

        method_options = (None, "nearest")
        if method not in method_options:
            raise ValueError(
                f"method must be one of {method_options}, received: {method}"
            )

        bounds_options = ("warn", "raise", "ignore", "drop")
        if out_of_bounds not in bounds_options:
            str_options = ", ".join(bounds_options)
            raise ValueError(
                f"out_of_bounds must be one of {str_options}, received: {out_of_bounds}"
            )

        # Check x and y coordinates
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        if x.shape != y.shape:
            raise ValueError("shape of x does not match shape of y")
        if x.ndim != 1:
            raise ValueError("x and y must be 1d")
        xy = np.column_stack([x, y])

        # Find which points are "in bounds" (i.e. not outside the topology)
        point_dim = f"{self.name}_points"
        core_indexer = self.locate_points(xy, tolerance)
        keep = slice(None, None)
        condition = None
        valid = core_indexer != -1
        if not valid.all():
            msg = "Not all points are located on the topology."
            if out_of_bounds == "raise":
                raise ValueError(msg)
            elif out_of_bounds == "warn":
                warnings.warn(msg, UserWarning, stacklevel=2)
                condition = xr.DataArray(valid, dims=(point_dim,))
            elif out_of_bounds == "ignore":
                condition = xr.DataArray(valid, dims=(point_dim,))
            else:  # "drop"
                core_indexer = core_indexer[valid]
                keep = valid
        xy_sel = xy[keep]

        # Create the selection DataArray or Dataset
        core_dim = self.core_dimension
        other_dims = self.dims.intersection(obj.dims) - {core_dim}
        facets = {v: k for k, v in self.facets.items()}
        # Override the main indexer if method is nearest.
        if core_dim in obj.dims:
            if method == "nearest":
                core_indexer = self._locate_nearest(
                    facet=facets[core_dim], points=xy_sel
                )
            indexers = {core_dim: xr.DataArray(core_indexer, dims=(point_dim,))}
        else:
            indexers = {}

        # Now process secondary dimensions.
        for dim in other_dims:
            indexer = self._locate_nearest(facet=facets[dim], points=xy_sel)
            indexers[dim] = xr.DataArray(indexer, dims=(point_dim,))

        selection = obj.isel(indexers).assign_coords(
            {
                f"{self.name}_x": (point_dim, xy[keep, 0]),
                f"{self.name}_y": (point_dim, xy[keep, 1]),
            }
        )
        # Set values to fill_value for out-of-bounds
        if condition is not None:
            # Utilize the exact name matching of .where() if our object is a dataset.
            if isinstance(selection, xr.Dataset):
                datavars = selection.data_vars.items()
                condition = xr.Dataset(
                    {
                        varname: condition
                        for varname, var in datavars
                        if point_dim in var.dims
                    }
                )
            selection = selection.where(condition, other=fill_value)
        return selection

    def locate_nearest_node(self, points: FloatArray, max_distance: float = np.inf):
        """
        Find which grid node is nearest for a collection of points.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``
        max_distance: optional, float

        Returns
        -------
        indices: ndarray of integers with shape ``(n_point,)``
            Missing indices are indicated with -1.
        """
        _, indices = self.node_kdtree.query(
            points, distance_upper_bound=max_distance, workers=-1
        )
        # The scipy KDTree returns missing indices (e.g. out of max_distance) with n.
        # We use -1 for consistency.
        indices[indices == self.n_node] = -1
        return indices

    def locate_nearest_edge(self, points: FloatArray, max_distance: float = np.inf):
        """
        Find which grid node is nearest for a collection of points.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``
        max_distance: optional, float

        Returns
        -------
        indices: ndarray of integers with shape ``(n_point,)``
            Missing indices are indicated with -1.
        """
        _, indices = self.edge_kdtree.query(
            points, distance_upper_bound=max_distance, workers=-1
        )
        # The scipy KDTree returns missing indices (e.g. out of max_distance) with n.
        # We use -1 for consistency.
        indices[indices == self.n_edge] = -1
        return indices

    def locate_points(self, points: FloatArray, tolerance: Optional[float] = None):
        """
        Find on which edge points are located.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. This
            accounts for the inherent inexactness of floating point calculations.
            If None, an appropriate tolerance is automatically estimated based on
            the geometry size. Consider adjusting this value if edge detection
            results are unsatisfactory.

        Returns
        -------
        edge_index: ndarray of integers with shape ``(n_points,)``
        """
        return self.celltree.locate_points(points, tolerance)

    def intersect_edges(self, edges: FloatArray):
        """
        Find in which grid edges are edges are located and compute the
        intersection with the edges.

        Parameters
        ----------
        edges: ndarray of floats with shape ``(n_edge, 2, 2)``
            The first dimensions represents the different edges.
            The second dimensions represents the start and end of every edge.
            The third dimensions reresent the x and y coordinate of every vertex.

        Returns
        -------
        edge_index: ndarray of integers with shape ``(n_intersection,)``
        edge_tree_index: ndarray of integers with shape ``(n_intersection,)``
        intersections: ndarray of float with shape ``(n_intersection, 2, 2)``
        """
        return self.celltree.intersect_edges(edges)

    def intersect_line(self, obj, start: Sequence[float], end: Sequence[float]):
        """
        Intersect a line with this grid, and fetch the values of the
        intersected faces.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset
        start: sequence of two floats
            coordinate pair (x, y), designating the start point of the line.
        end: sequence of two floats
            coordinate pair (x, y), designating the end point of the line.

        Returns
        -------
        selection: xr.DataArray or xr.Dataset
            The name of the topology is prefixed in the x, y and s
            (spatium=distance) coordinates.
        """
        if (len(start) != 2) or (len(end) != 2):
            raise ValueError("Start and end coordinate pairs must have length two")
        return self._sel_line(obj, start, end)

    def _sel_line(
        self,
        obj,
        start,
        end,
    ):
        dim = self.core_dimension
        edges = np.array([[start, end]])
        _, index, xy = self.intersect_edges(edges)
        coords, index = self._section_coordinates(edges, xy, dim, index, self.name)
        return obj.isel({dim: index}).assign_coords(coords)

    def _sel_yline(
        self,
        obj,
        x: slice,
        y: FloatArray,
    ):
        xmin, _, xmax, _ = self.bounds
        if y.size != 1:
            raise ValueError(
                "If x is a slice without steps, y should be a single value"
            )
        y = y[0]
        xstart = numeric_bound(x.start, xmin)
        xstop = numeric_bound(x.stop, xmax)
        return self._sel_line(obj, start=(xstart, y), end=(xstop, y))

    def _sel_xline(
        self,
        obj,
        x: FloatArray,
        y: slice,
    ):
        _, ymin, _, ymax = self.bounds
        if x.size != 1:
            raise ValueError(
                "If y is a slice without steps, x should be a single value"
            )
        x = x[0]
        ystart = numeric_bound(y.start, ymin)
        ystop = numeric_bound(y.stop, ymax)
        return self._sel_line(obj, start=(x, ystart), end=(x, ystop))

    def intersect_linestring(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
        linestring: "shapely.geometry.LineString",  # type: ignore # noqa
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Intersect linestrings with this grid, and fetch the values of the
        intersected faces.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset
        linestring: shapely.geometry.lineString

        Returns
        -------
        selection: xr.DataArray or xr.Dataset
            The name of the topology is prefixed in the x, y and s
            (spatium=distance) coordinates.
        """
        import shapely

        xy = shapely.get_coordinates([linestring])
        edges = np.stack((xy[:-1], xy[1:]), axis=1)
        edge_index, core_index, intersections = self.intersect_edges(edges)

        # Compute the cumulative length along the edges
        edge_length = np.linalg.norm(edges[:, 1] - edges[:, 0], axis=1)
        cumulative_length = np.empty_like(edge_length)
        cumulative_length[0] = 0
        np.cumsum(edge_length[:-1], out=cumulative_length[1:])

        # Compute the distance for every intersection to the start of the linestring.
        if self.topology_dimension == 2:
            intersection_for_coord = intersections.mean(axis=1)
        else:
            intersection_for_coord = intersections
        distance_node_to_intersection = np.linalg.norm(
            intersection_for_coord - edges[edge_index, 0], axis=1
        )
        s = distance_node_to_intersection + cumulative_length[edge_index]

        # Now sort everything according to s.
        dim = self.core_dimension
        coords, core_index = get_sorted_section_coords(
            s, intersection_for_coord, dim, core_index, self.name
        )

        return obj.isel({dim: core_index}).assign_coords(coords)

    def sel(self, obj, x=None, y=None):
        """
        Find selection in the UGRID x and y coordinates.

        The indexing for x and y always occurs orthogonally, i.e.:
        ``.sel(x=[0.0, 5.0], y=[10.0, 15.0])`` results in a four points. For
        vectorized indexing (equal to ``zip``ing through x and y), see
        ``.sel_points``.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset
        x: float, 1d array, slice
        y: float, 1d array, slice

        Returns
        -------
        dimension: str
        as_ugrid: bool
        index: 1d array of integers
        coords: dict
        """

        if x is None:
            x = slice(None, None)
        if y is None:
            y = slice(None, None)

        x = self._validate_indexer(x)
        y = self._validate_indexer(y)
        if isinstance(x, slice) and isinstance(y, slice):
            f = self._sel_box
        elif isinstance(x, slice) and isinstance(y, np.ndarray):
            f = self._sel_yline
        elif isinstance(x, np.ndarray) and isinstance(y, slice):
            f = self._sel_xline
        elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
            # Orthogonal points
            y, x = [a.ravel() for a in np.meshgrid(y, x, indexing="ij")]
            f = self.sel_points
        else:
            raise TypeError(
                f"Invalid indexer types: {type(x).__name__}, and {type(y).__name__}"
            )
        return f(obj, x, y)

    def _validate_partitioning_weights(self, weights: Optional[IntArray]) -> None:
        """Validate weights for partitioning. Check shape and type."""
        facet = {v: k for k, v in self.facets.items()}[self.core_dimension]
        n_expected = getattr(self, f"n_{facet}")
        if weights is None:
            return
        if weights.shape != (n_expected,):
            raise ValueError(
                f"Wrong shape on weights. Expected a 1D array with {n_expected} elements, "
                f"received array with shape: {weights.shape}"
            )
        if not np.issubdtype(weights.dtype, np.integer):
            raise TypeError(
                f"Wrong type on weights. Expected an integer array, received: {weights.dtype}"
            )
        if np.any(weights < 0):
            raise ValueError(
                "Wrong values on weights. Weights should be greater or equal to zero."
            )

    def label_partitions(self, n_part: int, weights: Optional[IntArray] = None):
        """
        Generate partition labels for this grid topology using METIS:
        https://github.com/KarypisLab/METIS

        This method utilizes the pymetis Python bindings:
        https://github.com/inducer/pymetis

        Parameters
        ----------
        n_part: integer
            The number of parts to partition the mesh.
        weights: optional, np.ndarray of integers
            The weight associated with each elements.

        Returns
        -------
        partition_labels: UgridDataArray of integers
        """
        import pymetis

        import xugrid

        self._validate_partitioning_weights(weights)
        facet = {v: k for k, v in self.facets.items()}[self.core_dimension]
        # E.g. node_node_connectivity (1D) or face_face_connectivity (2D)
        adjacency_matrix = getattr(self, f"{facet}_{facet}_connectivity")

        _, partition_index = pymetis.part_graph(
            nparts=n_part,
            adjacency=pymetis.CSRAdjacency(
                adj_starts=adjacency_matrix.indptr,
                adjacent=adjacency_matrix.indices,
            ),
            vweights=weights,
        )
        return xugrid.UgridDataArray(
            obj=xr.DataArray(
                data=np.array(partition_index),
                dims=(self.core_dimension,),
                name="labels",
            ),
            grid=self,
        )

    def partition(self, n_part: int, weights: Optional[IntArray] = None):
        """
        Partition this grid topology using METIS:
        https://github.com/KarypisLab/METIS

        This method utilizes the pymetis Python bindings:
        https://github.com/inducer/pymetis

        Parameters
        ----------
        n_part: integer
            The number of parts to partition the mesh.
        weights: optional, np.ndarray of integers
            The weight associated with each elements.

        Returns
        -------
        partitions
        """
        from xugrid.ugrid.partitioning import labels_to_indices

        labels = self.label_partitions(n_part, weights)
        indices = labels_to_indices(labels.values)
        return [self.topology_subset(index) for index in indices]


UgridType = Type[AbstractUgrid]



import numpy as np
import pandas as pd
import pytest
import xarray as xr

import xugrid as xu
from xugrid import (
    BarycentricInterpolator,
    CentroidLocatorRegridder,
    OverlapRegridder,
    RelativeOverlapRegridder,
)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_check_source_target_types(disk, cls):
    with pytest.raises(TypeError):
        cls(source=disk, target=1.0)
    with pytest.raises(TypeError):
        cls(source=1.0, target=disk)


@pytest.mark.parametrize(
    "regridder_class",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_structured_to_unstructured(
    regridder_class,
    disk,
    quads_structured,
):
    regridder = regridder_class(quads_structured, disk)
    actual = regridder.regrid(quads_structured)
    assert isinstance(actual, xu.UgridDataArray)

    # Regridding datasets is not (yet) allowed.
    ds = xr.Dataset({"a": quads_structured})
    with pytest.raises(TypeError):
        regridder.regrid(ds)


@pytest.mark.parametrize(
    "regridder_class",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_weights_as_dataframe(
    regridder_class,
    disk,
    quads_structured,
):
    regridder = regridder_class(quads_structured, disk)
    df = regridder.weights_as_dataframe()
    assert isinstance(df, pd.DataFrame)
    assert "source_index" in df
    assert "target_index" in df
    assert "weight" in df

    regridder._weights = None
    with pytest.raises(ValueError):
        regridder.weights_as_dataframe()


def test_centroid_locator_regridder_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_centroid
):
    regridder = CentroidLocatorRegridder(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result.fillna(0.0) == expected_results_centroid.fillna(0.0)).any()

    # With broadcasting
    regridder = CentroidLocatorRegridder(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (
        broadcasted.fillna(0.0).isel(layer=0) == expected_results_centroid.fillna(0.0)
    ).any()


def test_centroid_locator_regridder(disk, quads_1):
    square = quads_1
    regridder = CentroidLocatorRegridder(source=disk, target=square)
    result = regridder.regrid(disk)
    assert isinstance(result, xu.UgridDataArray)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()
    assert result.grid.n_face == square.grid.n_face

    # other way around
    regridder = CentroidLocatorRegridder(source=result, target=disk)
    back = regridder.regrid(result)
    assert isinstance(back, xu.UgridDataArray)
    assert back.notnull().any()
    assert back.min() >= disk.min()
    assert back.max() <= disk.max()
    assert back.grid.n_face == disk.grid.n_face

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * result.obj,
        result.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", disk.grid.face_dimension)


def test_overlap_regridder_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_overlap
):
    regridder = OverlapRegridder(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result == expected_results_overlap).any()

    # With broadcasting
    regridder = OverlapRegridder(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (broadcasted.isel(layer=0) == expected_results_overlap).any()

    # Test if "mode" method doesn't repeat first values again
    # https://github.com/Deltares/xugrid/issues/236
    grid_data_adapted = grid_data_a.copy()
    grid_data_adapted[0, 0] = 99
    regridder = OverlapRegridder(
        source=grid_data_adapted, target=grid_data_a, method="mode"
    )
    result = regridder.regrid(grid_data_adapted)
    assert not np.all(result == 99.0)


def test_overlap_regridder(disk, quads_1):
    square = quads_1
    regridder = OverlapRegridder(disk, square, method="mean")
    result = regridder.regrid(disk)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * disk.obj,
        grid=disk.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", square.grid.face_dimension)
    assert broadcasted.shape == (5, 100)


def test_linear_interpolator_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_linear
):
    regridder = BarycentricInterpolator(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result.fillna(0.0) == expected_results_linear.fillna(0.0)).any()

    # With broadcasting
    regridder = BarycentricInterpolator(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (
        broadcasted.fillna(0.0).isel(layer=0) == expected_results_linear.fillna(0.0)
    ).any()


def test_barycentric_interpolator(disk, quads_0_25):
    square = quads_0_25
    regridder = BarycentricInterpolator(source=disk, target=square)
    result = regridder.regrid(disk)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * disk.obj,
        grid=disk.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", square.grid.face_dimension)
    assert broadcasted.shape == (5, 1600)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_weights(cls, disk, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    weights = regridder.weights
    new_regridder = cls.from_weights(weights, target=square)
    new_result = new_regridder.regrid(disk)
    assert new_result.equals(result)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_weights_layered(cls, disk, disk_layered, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    weights = regridder.weights
    new_regridder = cls.from_weights(weights, target=square)
    new_result = new_regridder.regrid(disk_layered)
    assert np.array_equal(new_result.sel(layer=1).values, result.values, equal_nan=True)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_dataset(cls, disk, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    dataset = regridder.to_dataset()
    new_regridder = cls.from_dataset(dataset)
    new_result = new_regridder.regrid(disk)
    assert np.array_equal(new_result.values, result.values, equal_nan=True)


def test_regridder_daks_arrays(
    grid_data_dask_source,
    grid_data_dask_source_layered,
    grid_data_dask_target,
    grid_data_dask_expected,
    grid_data_dask_expected_layered,
):
    regridder = CentroidLocatorRegridder(
        source=grid_data_dask_source, target=grid_data_dask_target
    )
    result = regridder.regrid(grid_data_dask_source)
    assert result.equals(grid_data_dask_expected)

    # with broadcasting
    regridder = CentroidLocatorRegridder(
        source=grid_data_dask_source_layered, target=grid_data_dask_target
    )
    result = regridder.regrid(grid_data_dask_source_layered)
    assert result.isel(layer=0).equals(grid_data_dask_expected_layered.isel(layer=0))


def test_create_percentile_method():
    with pytest.raises(ValueError):
        OverlapRegridder.create_percentile_method(percentile=-1)

    with pytest.raises(ValueError):
        OverlapRegridder.create_percentile_method(percentile=101)

    median = OverlapRegridder.create_percentile_method(percentile=50)
    values = np.array([0, 1, 2, 3, 4], dtype=float)
    weights = np.ones_like(values)
    workspace = np.zeros_like(values)
    assert median(values, weights, workspace) == 2


def test_directional_dependence():
    # Increasing / decreasing x or y direction shouldn't matter for the result.
    da = xr.DataArray(
        data=[[1.0, 2.0], [3.0, 4.0]],
        coords={"y": [17.5, 12.5], "x": [2.5, 7.5]},
        dims=("y", "x"),
    )
    target_da = xr.DataArray(
        data=[[np.nan, np.nan], [np.nan, np.nan]],
        coords={"y": [10.0, 20.0], "x": [0.0, 10.0]},
        dims=("y", "x"),
    )
    target_uda = xu.UgridDataArray.from_structured2d(target_da)

    flip = slice(None, None, -1)
    flipy = da.isel(y=flip)
    flipx = da.isel(x=flip)
    flipxy = da.isel(x=flip, y=flip)
    uda = xu.UgridDataArray.from_structured2d(da)
    uda_flipxy = xu.UgridDataArray.from_structured2d(flipxy)

    # Structured target: test whether the result is the same regardless of source
    # orientation.
    result = []
    for source in [da, flipy, flipx, flipxy, uda, uda_flipxy]:
        regridder = xu.OverlapRegridder(source, target=target_da)
        result.append(regridder.regrid(source))
    first = result.pop(0)
    assert all(first.identical(item) for item in result)

    # Unstructured target: test whether the result is the same regardless of
    # source orientation.
    result = []
    for source in [da, flipy, flipx, flipxy, uda, uda_flipxy]:
        regridder = xu.OverlapRegridder(source, target=target_uda)
        result.append(regridder.regrid(source))
    first = result.pop(0)
    assert all(first.identical(item) for item in result)


def test_barycentric_concave():
    vertices = np.array(
        [
            [0.0, 0.0],
            [3.0, 0.0],
            [1.0, 1.0],
            [0.0, 2.0],
            [3.0, 2.0],
        ]
    )
    faces = np.array(
        [
            [0, 1, 2],
            [0, 2, 3],
            [2, 4, 3],
        ]
    )
    grid = xu.Ugrid2d(*vertices.T, -1, faces)

    dx = 0.1
    x = np.arange(0.0, 3.0, dx) + 0.5 * dx
    y = np.arange(0.0, 2.0, dx) + 0.5 * dx
    other = xr.DataArray(
        data=np.ones((y.size, x.size)), coords={"y": y, "x": x}, dims=("y", "x")
    )

    uda = xu.UgridDataArray(
        obj=xr.DataArray([2.0, 0.5, 2.0], dims=[grid.face_dimension]),
        grid=grid,
    )
    regridder = xu.BarycentricInterpolator(source=uda, target=other)
    result = regridder.regrid(uda)
    assert result.min() >= 0.5
    assert result.max() <= 2.0
    # There should be a large enough "gap" of NaNs on the right side.
    assert result.isnull().sum() == 200


def test_barycentric_structured():
    da = xr.DataArray(
        data=[[1, 2], [3, 4]],
        coords={"y": [1.5, 0.5], "x": [0.5, 1.5]},
        dims=("y", "x"),
    )
    x = np.arange(0.0, 2.0, 0.25) + 0.125
    y = np.arange(0.0, 2.0, 0.25) + 0.125
    target = xr.DataArray(
        data=np.empty((y.size, x.size)), coords={"y": y, "x": x}, dims=("y", "x")
    )

    regridder = xu.BarycentricInterpolator(source=da, target=target)
    out_structured = regridder.regrid(da)

    target_uda = xu.UgridDataArray.from_structured2d(target)
    regridder = xu.BarycentricInterpolator(source=da, target=target_uda)
    out_unstructured = regridder.regrid(da)

    # Answers should be same regardless whether the topology is stored as
    # structured or unstructured.
    values_structured = out_structured.to_numpy()
    values_unstructured = out_unstructured.to_numpy()
    assert np.allclose(values_structured.ravel(), values_unstructured)

    # Check values. Original values should be present at exterior (repeated).
    assert np.allclose(
        values_structured[0], [3.0, 3.0, 3.125, 3.375, 3.625, 3.875, 4.0, 4.0]
    )
    assert np.allclose(
        values_structured[-1], [1.0, 1.0, 1.125, 1.375, 1.625, 1.875, 2.0, 2.0]
    )
    assert np.allclose(
        values_structured[:, 0], [3.0, 3.0, 2.75, 2.25, 1.75, 1.25, 1.0, 1.0]
    )



import numpy as np

from xugrid.regrid import utils


def test_create_linear_index():
    index_a = [0, 0, 1]
    index_b = [0]
    shape = (2, 1)
    actual = utils.create_linear_index((index_a, index_b), shape)
    expected = np.array([0, 0, 1])
    assert np.array_equal(actual, expected)

    index_a = [0, 0, 1, 1]
    index_b = [0, 1, 2]
    shape = (2, 3)
    actual = utils.create_linear_index((index_a, index_b), shape)
    expected = np.array([0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5])
    assert np.array_equal(actual, expected)


def test_create_weights():
    weights_a = [0.25, 0.25, 0.25, 0.25]
    weights_b = [0.5, 0.5]
    actual = utils.create_weights((weights_a, weights_b))
    expected = np.full(8, 0.125)
    assert np.allclose(actual, expected)


def test_broadcast():
    source_shape = (3, 2)
    target_shape = (6, 4)
    index_src_a = [0, 0, 1, 1]
    index_src_b = [0, 1]
    index_tgt_a = [2, 2, 3, 3]
    index_tgt_b = [2, 3]
    weights_a = [0.5, 0.5, 0.5, 0.5]
    weights_b = [0.5, 0.5]
    actual_src, actual_tgt, actual_weights = utils.broadcast(
        source_shape,
        target_shape,
        (index_src_a, index_src_b),
        (index_tgt_a, index_tgt_b),
        (weights_a, weights_b),
    )
    expected_src = np.array([0, 1, 0, 1, 2, 3, 2, 3])
    expected_tgt = np.array([10, 11, 10, 11, 14, 15, 14, 15])
    expected_weights = np.full(8, 0.25)
    assert np.array_equal(actual_src, expected_src)
    assert np.array_equal(actual_tgt, expected_tgt)
    assert np.allclose(actual_weights, expected_weights)


def test_alt_cumsum():
    a = np.ones(5)
    assert np.array_equal(utils.alt_cumsum(a), np.arange(5))

    a = np.array([1, 3, 4])
    assert np.array_equal(utils.alt_cumsum(a), [0, 1, 4])



"""Create and merge partitioned UGRID topologies."""

from collections import defaultdict
from itertools import accumulate, chain
from typing import List

import numpy as np
import xarray as xr

from xugrid.constants import FILL_VALUE, IntArray, IntDType
from xugrid.core.wrap import UgridDataArray, UgridDataset
from xugrid.ugrid.connectivity import renumber
from xugrid.ugrid.ugridbase import UgridType


def labels_to_indices(labels: IntArray) -> List[IntArray]:
    """
    Convert a 1D array of N labels into a N arrays of indices.

    E.g. [0, 1, 0, 2, 2] -> [[0, 2], [1], [3, 4]]
    """
    sorter = np.argsort(labels)
    split_indices = np.cumsum(np.bincount(labels)[:-1])
    indices = np.split(sorter, split_indices)
    for index in indices:
        index.sort()
    return indices


def partition_by_label(grid, obj, labels: IntArray):
    """
    Partition the grid and xarray object by integer labels.

    This function is used by UgridDataArray.partition_by_label and
    UgridDataset.partition_by_label.

    Parameters
    ----------
    grid: Ugrid1d, Ugrid2d
    obj: DataArray or Dataset
    labels: UgridDataArray of integers

    Returns
    -------
    partitions: List of (grid, obj)
    """
    if not isinstance(labels, UgridDataArray):
        raise TypeError(
            f"labels must be a UgridDataArray, received: {type(labels).__name__}"
        )
    if not np.issubdtype(labels.dtype, np.integer):
        raise TypeError(f"labels must have integer dtype, received {labels.dtype}")

    if labels.grid != grid:
        raise ValueError("grid of labels does not match xugrid object")
    if labels.dims != (grid.core_dimension,):
        raise ValueError(
            f"Can only partition this topology by {grid.core_dimension}, found"
            f" the dimensions: {labels.dims}"
        )

    if isinstance(obj, xr.Dataset):
        obj_type = UgridDataset
    elif isinstance(obj, xr.DataArray):
        obj_type = UgridDataArray
    else:
        raise TypeError(
            f"Expected DataArray or Dataset, received: {type(obj).__name__}"
        )

    indices = labels_to_indices(labels.values)
    partitions = []
    for index in indices:
        new_grid, indexes = grid.topology_subset(index, return_index=True)
        new_obj = obj.isel(indexes, missing_dims="ignore")
        partitions.append(obj_type(new_obj, new_grid))

    return partitions


def merge_nodes(grids):
    node_x = np.hstack([grid.node_x for grid in grids])
    node_y = np.hstack([grid.node_y for grid in grids])
    node_xy = np.column_stack((node_x, node_y))
    _, index, inverse = np.unique(
        node_xy, axis=0, return_index=True, return_inverse=True
    )
    inverse = inverse.ravel()
    # We want to maintain order, so create an inverse index to the new numbers.
    inverse = renumber(index)[inverse]
    # Maintain order.
    index.sort()
    unique_nodes = node_xy[index]
    # Slice the indexes per partition.
    slices = (0,) + tuple(accumulate(grid.n_node for grid in grids))
    sections = np.searchsorted(index, slices[1:-1])
    indexes = np.split(index, sections)
    for partition_index, offset in zip(indexes, slices):
        partition_index -= offset
    return unique_nodes, indexes, inverse


def _merge_connectivity(gathered, slices):
    # Make sure identical edges are identified: [0, 1] == [1, 0]
    # As well as faces: [0, 1, 2] == [2, 1, 0]
    sorted = np.sort(gathered, axis=1)
    _, index = np.unique(sorted, axis=0, return_index=True)
    # Maintain order.
    index.sort()
    merged = gathered[index]
    # Slice the indexes per partition.
    sections = np.searchsorted(index, slices[1:-1])
    indexes = np.split(index, sections)
    for partition_index, offset in zip(indexes, slices):
        partition_index -= offset
    return merged, indexes


def merge_faces(grids, node_inverse):
    node_offsets = tuple(accumulate([0] + [grid.n_node for grid in grids]))
    n_face = [grid.n_face for grid in grids]
    n_max_node = max(grid.n_max_node_per_face for grid in grids)
    slices = (0,) + tuple(accumulate(n_face))

    all_faces = np.full((sum(n_face), n_max_node), FILL_VALUE, dtype=IntDType)
    for grid, face_offset, node_offset in zip(grids, slices, node_offsets):
        faces = grid.face_node_connectivity
        n_face, n_node_per_face = faces.shape
        valid = faces != FILL_VALUE
        all_faces[face_offset : face_offset + n_face, :n_node_per_face][valid] = (
            node_inverse[faces[valid] + node_offset]
        )

    return _merge_connectivity(all_faces, slices)


def merge_edges(grids, node_inverse):
    node_offsets = tuple(accumulate([0] + [grid.n_node for grid in grids]))
    n_edge = [grid.n_edge for grid in grids]
    slices = (0,) + tuple(accumulate(n_edge))

    all_edges = np.empty((sum(n_edge), 2), dtype=IntDType)
    for grid, edge_offset, offset in zip(grids, slices, node_offsets):
        edges = grid.edge_node_connectivity
        n_edge = len(edges)
        all_edges[edge_offset : edge_offset + n_edge] = node_inverse[edges + offset]

    return _merge_connectivity(all_edges, slices)


def validate_partition_topology(grouped: defaultdict[str, UgridType]) -> None:
    for name, grids in grouped.items():
        types = {type(grid) for grid in grids}
        if len(types) > 1:
            raise TypeError(
                f"All partition topologies with name {name} should be of the "
                f"same type, received: {types}"
            )

        griddims = list({tuple(grid.dims) for grid in grids})
        if len(griddims) > 1:
            raise ValueError(
                f"Dimension names on UGRID topology {name} do not match "
                f"across partitions: {griddims[0]} versus {griddims[1]}"
            )

    return None


def group_grids_by_name(partitions: list[UgridDataset]) -> defaultdict[str, UgridType]:
    grouped = defaultdict(list)
    for partition in partitions:
        for grid in partition.grids:
            grouped[grid.name].append(grid)

    validate_partition_topology(grouped)
    return grouped


def group_data_objects_by_gridname(
    partitions: list[UgridDataset],
) -> defaultdict[str, xr.Dataset]:
    # Convert to dataset for convenience
    data_objects = [partition.obj for partition in partitions]
    data_objects = [
        obj.to_dataset() if isinstance(obj, xr.DataArray) else obj
        for obj in data_objects
    ]

    grouped = defaultdict(list)
    for partition, obj in zip(partitions, data_objects):
        for grid in partition.grids:
            grouped[grid.name].append(obj)

    return grouped


def validate_partition_objects(
    objects_by_gridname: defaultdict[str, xr.Dataset],
) -> None:
    for data_objects in objects_by_gridname.values():
        allvars = list({tuple(sorted(ds.data_vars)) for ds in data_objects})
        unique_vars = set(chain(*allvars))
        # Check dimensions
        dims_per_var = [
            {ds.variables[var].dims for ds in data_objects if var in ds.data_vars}
            for var in unique_vars
        ]
        for var, vardims in zip(unique_vars, dims_per_var):
            if len(vardims) > 1:
                vardims_ls = list(vardims)
                raise ValueError(
                    f"Dimensions for '{var}' do not match across partitions: "
                    f"{vardims_ls[0]} versus {vardims_ls[1]}"
                )
    return None


def separate_variables(
    objects_by_gridname: defaultdict[str, xr.Dataset], ugrid_dims: set[str]
):
    """Separate into UGRID variables grouped by dimension, and other variables."""
    validate_partition_objects(objects_by_gridname)

    def assert_single_dim(intersection):
        if len(intersection) > 1:
            raise ValueError(
                f"{var} contains more than one UGRID dimension: {intersection}"
            )

    def remove_item(tuple, index):
        return tuple[:index] + tuple[index + 1 :]

    def all_equal(iterator):
        first = next(iter(iterator))
        return all(element == first for element in iterator)

    # Group variables by UGRID dimension.
    grouped = defaultdict(set)  # UGRID associated vars
    other = defaultdict(set)  # other vars

    for gridname, data_objects in objects_by_gridname.items():
        variables = {
            varname: data
            for obj in data_objects
            for varname, data in obj.variables.items()
        }
        vardims = {varname: data.dims for varname, data in variables.items()}
        for var, dims in vardims.items():
            shapes = [obj[var].shape for obj in data_objects if var in obj]

            # Check if variable depends on UGRID dimension.
            intersection = ugrid_dims.intersection(dims)
            if intersection:
                assert_single_dim(intersection)
                # Now check whether the non-UGRID dimensions match.
                dim = intersection.pop()  # Get the single element in the set.
                axis = dims.index(dim)
                shapes = [remove_item(shape, axis) for shape in shapes]
                if all_equal(shapes):
                    grouped[dim].add(var)

            elif all_equal(shapes):
                other[gridname].add(var)

    return grouped, other


def merge_data_along_dim(
    data_objects: list[xr.Dataset],
    vars: list[str],
    merge_dim: str,
    indexes: list[np.array],
    merged_grid: UgridType,
) -> xr.Dataset:
    """
    Select variables from the data objects.
    Pad connectivity dims if needed.
    Concatenate along dim.
    """
    max_sizes = merged_grid.max_connectivity_sizes
    ugrid_connectivity_dims = set(max_sizes)

    to_merge = []
    for obj, index in zip(data_objects, indexes):
        # Check for presence of vars
        missing_vars = set(vars).difference(set(obj.variables.keys()))
        if missing_vars:
            raise ValueError(f"Missing variables: {missing_vars} in partition {obj}")

        selection = obj[vars].isel({merge_dim: index}, missing_dims="ignore")

        # Pad the ugrid connectivity dims (e.g. n_max_face_node_connectivity) if
        # needed.
        present_dims = ugrid_connectivity_dims.intersection(selection.dims)
        pad_width = {}
        for dim in present_dims:
            nmax = max_sizes[dim]
            size = selection.sizes[dim]
            if size != nmax:
                pad_width[dim] = (0, nmax - size)
        if pad_width:
            selection = selection.pad(pad_width=pad_width)

        to_merge.append(selection)

    return xr.concat(to_merge, dim=merge_dim)


def single_ugrid_chunk(da: xr.DataArray, ugrid_dims: set[str]) -> xr.DataArray:
    """
    Ensure that along a UGRID dimension only a single chunk is defined.

    Preserving chunks on the merged result may generate a very sub-optimal,
    complex dask graph.
    """
    if not da.chunks:
        return da

    chunks = {}
    for dim, sizes in zip(da.dims, da.chunks):
        # Define a single chunk for each UGRID dimension.
        if dim in ugrid_dims:
            chunks[dim] = (da.sizes[dim],)
        else:
            chunks[dim] = sizes
    return da.chunk(chunks)


def merge_partitions(partitions, merge_ugrid_chunks: bool = True):
    """
    Merge topology and data, partitioned along UGRID dimensions, into a single
    UgridDataset.

    UGRID topologies and variables are merged if they share a name. Topologies
    and variables must be present in *all* partitions. Dimension names must
    match.

    Variables are omitted from the merged result if non-UGRID dimensions do not
    match in size.

    Parameters
    ----------
    partitions : sequence of UgridDataset or UgridDataArray
    merge_ugrid_chunks: bool, default is True.
        Whether to merge chunks along the UGRID topology dimensions.

    Returns
    -------
    merged : UgridDataset
    """
    if len(partitions) == 0:
        raise ValueError("Cannot merge partitions: zero partitions provided.")
    types = {type(obj) for obj in partitions}
    msg = "Expected UgridDataArray or UgridDataset, received: {}"
    if len(types) > 1:
        type_names = [t.__name__ for t in types]
        raise TypeError(msg.format(type_names))
    obj_type = types.pop()
    if obj_type not in (UgridDataArray, UgridDataset):
        raise TypeError(msg.format(obj_type.__name__))

    # return first partition if single partition is provided
    if len(partitions) == 1:
        return next(iter(partitions))

    # Collect grids
    grids = [grid for p in partitions for grid in p.grids]
    ugrid_dims = {dim for grid in grids for dim in grid.dims}
    grids_by_name = group_grids_by_name(partitions)

    data_objects_by_name = group_data_objects_by_gridname(partitions)
    vars_by_dim, other_vars_by_name = separate_variables(
        data_objects_by_name, ugrid_dims
    )

    # First, take identical non-UGRID variables from the first partition:
    merged = xr.Dataset()

    # Merge the UGRID topologies into one, and find the indexes to index into
    # the data to avoid duplicates.
    merged_grids = []
    for gridname, grids in grids_by_name.items():
        data_objects = data_objects_by_name[gridname]
        other_vars = other_vars_by_name[gridname]

        # First, merge the grid topology.
        grid = grids[0]
        merged_grid, indexes = grid.merge_partitions(grids)
        merged_grids.append(merged_grid)

        # Add other vars, unassociated with UGRID dimensions, to dataset.
        for obj in data_objects:
            other_vars_obj = set(other_vars).intersection(set(obj.data_vars))
            merged.update(obj[other_vars_obj])

        for dim, dim_indexes in indexes.items():
            vars = vars_by_dim[dim]
            if len(vars) == 0:
                continue
            merged_selection = merge_data_along_dim(
                data_objects, vars, dim, dim_indexes, merged_grid
            )
            merged.update(merged_selection)

    # Merge chunks along the UGRID dimensions
    if merge_ugrid_chunks:
        # Running `merged[varname] = da` will cause xarray to auto-align. This
        # is quite expensive, especially when there are many variables. Setting
        # it via `._variables` sidesteps the alignment.
        for varname, da in merged._variables.items():
            merged._variables[varname] = single_ugrid_chunk(da, ugrid_dims)

    return UgridDataset(merged, merged_grids)



from typing import Callable, Union

import xarray as xr

import xugrid
from xugrid.core.sparse import MatrixCSR
from xugrid.regrid import reduce
from xugrid.regrid.network import Network1d
from xugrid.regrid.regridder import BaseRegridder, make_regrid, setup_grid
from xugrid.regrid.structured import StructuredGrid2d
from xugrid.regrid.unstructured import UnstructuredGrid2d


def convert_to_match(source, target):
    PROMOTIONS = {
        frozenset({StructuredGrid2d}): UnstructuredGrid2d,
        frozenset({UnstructuredGrid2d}): UnstructuredGrid2d,
    }
    types = set({type(target)})
    matched_type = PROMOTIONS[frozenset(types)]
    return source, target.convert_to(matched_type)


class NetworkGridder(BaseRegridder):
    """
    Network gridder for 2D unstructured grids.

    Parameters
    ----------
    grid: Ugrid1d
        The grid to be used for the regridding.
    """

    _JIT_FUNCTIONS = {
        k: make_regrid(f) for k, f in reduce.ABSOLUTE_OVERLAP_METHODS.items()
    }

    def __init__(
        self,
        source: "xugrid.Ugrid1d",
        target: "xugrid.Ugrid2d",
        method: Union[str, Callable] = "mean",
    ):
        self._source = Network1d(source)
        self._target = setup_grid(target)
        self._weights = None
        self._compute_weights(self._source, self._target, relative=False)
        self._setup_regrid(method)
        return

    @property
    def weights(self):
        return self.to_dataset()

    @weights.setter
    def weights(self, weights: MatrixCSR):
        if not isinstance(weights, MatrixCSR):
            raise TypeError(f"Expected MatrixCSR, received: {type(weights).__name__}")
        self._weights = weights
        return

    @classmethod
    def _weights_from_dataset(cls, dataset: xr.Dataset) -> MatrixCSR:
        return cls._csr_from_dataset(dataset)

    def _compute_weights(self, source, target, relative: bool) -> None:
        source, target = convert_to_match(source, target)
        source_index, target_index, weight_values = target.intersection_length(
            source, relative=relative
        )
        self._weights = MatrixCSR.from_triplet(
            target_index, source_index, weight_values, n=target.size, m=source.size
        )
        return

    @classmethod
    def from_weights(
        cls,
        weights: xr.Dataset,
        target: Union["xugrid.Ugrid2d", xr.DataArray, xr.Dataset],
        method: Union[str, Callable] = "mean",
    ):
        instance = super().from_weights(weights, target)
        instance._setup_regrid(method)
        return instance
